import copy

try:
    from osgeo import ogr
    from osgeo import osr
    import_ogr_success = True
except ImportError:
    import_ogr_success = False

from geoformat.conf.error_messages import geometry_must_be_polygon_or_multipolygon
from geoformat.conversion.bbox_conversion import bbox_extent_to_2d_bbox_extent
from geoformat.conversion.coordinates_conversion import (
    coordinates_to_2d_coordinates,
    format_coordinates,
    force_rhr_polygon_coordinates
)
from geoformat.conf.geometry_variable import GEOMETRY_CODE_TO_GEOMETRY_TYPE
from geoformat.conversion.bytes_conversion import big_endian_wkb_geometry_type_to_geometry_type, \
    geometry_type_to_wkb_geometry_type, int_to_4_bytes_integer, coordinates_list_to_bytes, integer_4_bytes_to_int, \
    double_8_bytes_to_float
from geoformat.geoprocessing.connectors.operations import coordinates_to_bbox
from geoformat.geoprocessing.geoparameters.bbox import bbox_union

from geoformat.conf.error_messages import import_ogr_error

geometry_type_to_dimension = {
    'Point': 2,
    'LineString': 2,
    'Polygon': 2,
    'MultiPoint': 2,
    'MultiLineString': 2,
    'MultiPolygon': 2,
}

geometry_type_to_upper_case = {
    'Point': 'POINT',
    'LineString': 'LINESTRING',
    'Polygon': 'POLYGON',
    'MultiPoint': 'MULTIPOINT',
    'MultiLineString': 'MULTILINESTRING',
    'MultiPolygon': 'MULTIPOLYGON',
    'GeometryCollection': 'GEOMETRYCOLLECTION'
}
upper_case_to_geometry_type = {
    'POINT': 'Point',
    'LINESTRING': 'LineString',
    'POLYGON': 'Polygon',
    'MULTIPOINT': 'MultiPoint',
    'MULTILINESTRING': 'MultiLineString',
    'MULTIPOLYGON': 'MultiPolygon',
    'GEOMETRYCOLLECTION': 'GeometryCollection'
}


def geometry_type_to_2d_geometry_type(geometry_type):
    """Convert xD geometry type to 2D geometry type

    :param geometry_type
    :return 2d geometry type
    """
    if 'POINT' in geometry_type.upper():
        new_geometry_type = 'Point'
    elif 'LINESTRING' in geometry_type.upper():
        new_geometry_type = 'LineString'
    elif 'POLYGON' in geometry_type.upper():
        new_geometry_type = 'Polygon'
    elif 'GEOMETRYCOLLECTION' in geometry_type.upper():
        new_geometry_type = 'GeometryCollection'
    else:
        raise Exception("Geometry type unknown")

    if 'MULTI' in geometry_type.upper():
        new_geometry_type = 'Multi' + new_geometry_type

    return new_geometry_type


def geometry_to_2d_geometry(geometry, bbox=True):
    """
    Convert a geometry with 2 or more dimension to 2d dimension geometry.

    :param geometry: input geometry with 2 or more dimension
    :param bbox: boolean that indicate if we want or not the bbox information in geometry
    :return new geometry with 2 dimensions
    """
    geometry_collection = geometry_to_geometry_collection(geometry, bbox=False)
    for i_geom, geom in enumerate(geometry_collection['geometries']):
        new_geometry_type = geometry_type_to_2d_geometry_type(geom['type'])
        new_geometry = {'type': new_geometry_type, 'coordinates': coordinates_to_2d_coordinates(geom['coordinates'])}
        if bbox:
            if 'bbox' in geom:
                bbox = bbox_extent_to_2d_bbox_extent(geom['bbox'])
            else:
                bbox = coordinates_to_bbox(new_geometry['coordinates'])

            new_geometry['bbox'] = bbox

        geometry_collection['geometries'][i_geom] = new_geometry

    if 'GeometryCollection' in geometry['type']:
        # in case of geometry == GeometryCollection25D
        if geometry['type'] != 'GeometryCollection':
            geometry_collection['type'] = 'GeometryCollection'
        return geometry_collection
    else:
        return geometry_collection['geometries'][0]


def geometry_to_geometry_collection(geometry, geometry_type_filter=None, bbox=True):
    """
    Transform a geometry to GeometryCollection
    """
    if geometry_type_filter:
        if isinstance(geometry_type_filter, str):
            geometry_type_filter = {geometry_type_filter}
        elif isinstance(geometry_type_filter, (list, tuple, set)):
            geometry_type_filter = set(geometry_type_filter)
        else:
            raise Exception('geometry_type_filter must be a geometry type or a list of geometry type')

    geometry_type = geometry['type']
    # if input geometry is not an geometry collection
    if 'GeometryCollection' not in geometry_type:
        geometry_in_collection = copy.deepcopy(geometry)
        if geometry_type_filter:
            if geometry_type in geometry_type_filter:
                geometry_collection = {'type': 'GeometryCollection', 'geometries': [geometry_in_collection]}
            else:
                geometry_collection = {'type': 'GeometryCollection', 'geometries': []}
        else:
            geometry_collection = {'type': 'GeometryCollection', 'geometries': [geometry_in_collection]}

        if bbox and geometry['coordinates']:
            if 'bbox' in geometry:
                bbox = geometry['bbox']
            else:
                bbox = coordinates_to_bbox(geometry['coordinates'])
                geometry_in_collection['bbox'] = bbox
            geometry_collection['bbox'] = bbox

    # if input geometry is a geometry collection
    else:
        # if geometry is not empty
        if geometry['geometries']:
            if bbox:
                bbox_geometry_collection = None

            geometry_list = [None] * len(geometry['geometries'])
            del_idx = []
            for i_geom, geometry in enumerate(geometry['geometries']):
                geometry_in_collection = copy.deepcopy(geometry)
                write_geometry = True
                if geometry_type_filter:
                    if geometry['type'] not in geometry_type_filter:
                        write_geometry = False
                else:
                    if bbox:
                        if 'bbox' not in geometry:
                            bbox_geometry = coordinates_to_bbox(geometry_in_collection['coordinates'])

                            if not bbox_geometry_collection and bbox_geometry:
                                bbox_geometry_collection = bbox_geometry
                            else:
                                bbox_geometry_collection = bbox_union(bbox_geometry, bbox_geometry_collection)

                            if bbox and bbox_geometry:
                                geometry_in_collection['bbox'] = bbox_geometry

                # write geometry in list
                if write_geometry:
                    geometry_list[i_geom] = geometry_in_collection
                else:
                    del_idx.append(i_geom)

            # delete filter geometries
            if del_idx:
                for idx in reversed(del_idx):
                    del geometry_list[idx]

            geometry_collection = {'type': geometry_type, 'geometries': geometry_list}

            if bbox:
                if bbox_geometry_collection:
                    geometry_collection['bbox'] = bbox_geometry_collection
        else:
            geometry_collection = {'type': geometry_type, 'geometries': []}

    return geometry_collection


def single_geometry_to_multi_geometry(geometry, bbox=False):
    """
    Transform a single geometry (Point, Linestring or Polygon) to multi geometry (MultiPoint, MultiLineString,
    MultiPolygon).

    Equivalent to st_multi() in postgis

    :param geometry: geometry to transform to Multi geometry
    :return: mutli geometry
    """
    original_type = geometry['type']

    geometry_collection = geometry_to_geometry_collection(geometry=geometry, bbox=bbox)
    output_geometry_list = geometry_collection['geometries']

    for geometry_in_collection in output_geometry_list:
        if not 'MULTI' in geometry_in_collection['type'].upper():
            geometry_in_collection['type'] = 'Multi' + geometry_in_collection['type']
            geometry_in_collection['coordinates'] = [geometry_in_collection['coordinates']]

    if original_type.upper() != 'GEOMETRYCOLLECTION':
        output_geometry = output_geometry_list[0]
    else:
        output_geometry = geometry_collection

    return output_geometry


def multi_geometry_to_single_geometry(geometry, bbox=False):
    """
    Iterator in given geometry and send single geometry (Point, LineString, Polygon) if geometry is a multigeometry.
    Works with GeometryCollection
    """

    if geometry['type'].upper() == 'GEOMETRYCOLLECTION':
        for inside_geometry in geometry['geometries']:
            for single_geom in multi_geometry_to_single_geometry(inside_geometry, bbox=bbox):
                if bbox is True:
                    single_geom['bbox'] = coordinates_to_bbox(single_geom['coordinates'])
                yield single_geom

    elif geometry['type'].upper() in {'MULTIPOINT', 'MULTILINESTRING', 'MULTIPOLYGON'}:
        if geometry['type'].upper() == 'MULTIPOINT':
            single_geometry_type = 'Point'
        elif geometry['type'].upper() == 'MULTILINESTRING':
            single_geometry_type = 'LineString'
        else:
            single_geometry_type = 'Polygon'

        multi_coordinates = geometry['coordinates']
        for coordinates in multi_coordinates:
            output_geometry = {'type': single_geometry_type, 'coordinates': coordinates}
            if bbox is True:
                output_geometry['bbox'] = coordinates_to_bbox(coordinates)
            yield output_geometry

    else:
        output_geometry = copy.deepcopy(geometry)
        if bbox is True:
            if 'bbox' not in geometry:
                output_geometry['bbox'] = coordinates_to_bbox(output_geometry['coordinates'])
        yield output_geometry


def geometry_to_multi_geometry(geometry, bbox=True):
    """
    Convert a single geometry to multi geometry.

    :param geometry: input geometry
    :param bbox: boolean that indicate if we want or not the bbox information in geometry
    :return: geometry transform to multi geometry
    """

    if 'GEOMETRYCOLLECTION' in geometry['type'].upper():
        output_geometry = {"type": "GeometryCollection"}
        geometries_list = [None] * len(geometry["geometries"])
        compute_bbox = False
        if bbox is True:
            if "bbox" in geometry:
                compute_bbox = False
                collection_bbox = geometry["bbox"]
            else:
                compute_bbox = True

        for i_geom, geometry_in_collection in enumerate(geometry["geometries"]):
            geometry_in_multi = geometry_to_multi_geometry(geometry_in_collection, bbox=bbox)
            geometries_list[i_geom] = geometry_in_multi
            if compute_bbox is True:
                if i_geom == 0:
                    collection_bbox = geometry_in_multi["bbox"]
                else:
                    collection_bbox = bbox_union(collection_bbox, geometry_in_multi["bbox"])

        output_geometry["geometries"] = geometries_list

        if bbox is True:
            output_geometry["bbox"] = collection_bbox
    else:
        if 'MULTI' in geometry['type'].upper():
            output_geometry = copy.deepcopy(geometry)
        else:
            output_geometry = {"type": "Multi" + geometry["type"], "coordinates": [geometry["coordinates"]]}

        if bbox is True:
            if 'bbox' in geometry:
                output_geometry['bbox'] = geometry['bbox']
            else:
                output_geometry['bbox'] = coordinates_to_bbox(geometry['coordinates'])

    return output_geometry


def ogr_geometry_to_geometry(ogr_geometry, bbox=True):
    """
    Convert GDAL/OGR geometry to geoformat geometry

    :param ogr_geometry: ogr geometry class
    :param bbox: boolean that indicate if we want or not the bbox information in geometry
    :return: geoformat geometry
    """

    def coordinates_loop(ogr_geometry, bbox_launch):
        # if geometry is empty
        if ogr_geometry.IsEmpty() is True:
            return [], None
        else:
            bbox = ()
            # if geometry collection
            if ogr_geometry.GetGeometryType() == 7:
                geom_collection_list = []
                for i, under_geom in enumerate(ogr_geometry):
                    geom_geojson = ogr_geometry_to_geometry(
                        ogr_geometry=under_geom,
                        bbox=bbox_launch
                    )
                    geom_collection_list.append(geom_geojson)
                    if bbox_launch and 'bbox' in geom_geojson:
                        if bbox == ():
                            bbox = geom_geojson.get('bbox')
                        else:
                            bbox = bbox_union(bbox, geom_geojson['bbox'])

                return geom_collection_list, tuple(bbox)

            # if point
            elif ogr_geometry.GetGeometryType() == 1:
                coordinates = [ogr_geometry.GetX(), ogr_geometry.GetY()]
                bbox = (ogr_geometry.GetX(), ogr_geometry.GetY(), ogr_geometry.GetX(), ogr_geometry.GetY())

                return coordinates, bbox

            # for linestring, polygon, multipoint, multilinestring, multipolygon
            else:
                # iterate over geometries
                if ogr_geometry.GetGeometryCount() > 0:
                    coordinates_list = [0] * ogr_geometry.GetGeometryCount()
                    for j, under_geom in enumerate(ogr_geometry):
                        # linearring compose polygon if j > 0 then j is a hole.
                        # Hole is inside polygon so we do not need to loop on this coordinate
                        if under_geom.GetGeometryName() == 'LINEARRING' and j > 0:
                            bbox_launch = False
                        else:
                            bbox_launch = True

                        coordinates_list[j], under_bbox = coordinates_loop(under_geom, bbox_launch)
                        if bbox_launch:
                            # bbox
                            if bbox == () and under_bbox:
                                bbox = under_bbox
                            else:
                                bbox = bbox_union(bbox, under_bbox)

                    return coordinates_list, tuple(bbox)

                # iterate over points
                if ogr_geometry.GetPointCount() > 0:
                    coordinates_list = [0] * ogr_geometry.GetPointCount()
                    for i in range(ogr_geometry.GetPointCount()):
                        # coordinates
                        point = ogr_geometry.GetPoint(i)
                        coordinates_list[i] = [point[0], point[1]]
                        # bbox
                        if bbox_launch:
                            if i == 0:
                                bbox = [point[0], point[1], point[0], point[1]]
                            else:
                                if point[0] < bbox[0]:
                                    bbox[0] = point[0]
                                if point[0] > bbox[2]:
                                    bbox[2] = point[0]
                                if point[1] < bbox[1]:
                                    bbox[1] = point[1]
                                if point[1] > bbox[3]:
                                    bbox[3] = point[1]

                    return coordinates_list, tuple(bbox)

    if import_ogr_success:
        geometry = {}
        ogr_geom_type = ogr_geometry.GetGeometryType()
        geometry['type'] = GEOMETRY_CODE_TO_GEOMETRY_TYPE[ogr_geom_type]

        geometry_coordinates, geometry_bbox = coordinates_loop(
            ogr_geometry=ogr_geometry,
            bbox_launch=bbox
        )

        # adding bbox
        if geometry_coordinates and bbox is True:
            geometry['bbox'] = geometry_bbox

        # adding coordinates for geometries
        # if GeometryCollection
        if ogr_geom_type != 7:
            geometry['coordinates'] = geometry_coordinates
        # for others geometries
        else:
            geometry['geometries'] = geometry_coordinates

        return geometry
    else:
        raise Exception(import_ogr_error)


def geometry_to_ogr_geometry(geometry):
    """
    Convert geometry to GDAL/OGR geometry

    :param geometry: input geometry
    :return: ogr geometry object
    """

    def coordinates_loop(coordinates, ogr_geom):
        """
        2D ONLY -- if 3D change AddPoint_2D by AddPoint and add 3D type in GetGeometryType test.
        :param coordinates:
        :param ogr_geom:
        :return:
        """

        if coordinates:
            # if geometry collection
            if ogr_geom.GetGeometryType() == 7:
                for geojson_geom in coordinates:
                    new_ogr_geom = geometry_to_ogr_geometry(geojson_geom)
                    ogr_geom.AddGeometry(new_ogr_geom)

                return ogr_geom

            # if point
            elif ogr_geom.GetGeometryType() == 1:
                ogr_geom.AddPoint_2D(coordinates[0], coordinates[1])
                return ogr_geom

            # for linestring, polygon, multipoint, multilinestring, multipolygon
            else:
                if ogr_geom.GetGeometryType() == 4:
                    under_ogr_geom = ogr.Geometry(ogr.wkbPoint)
                elif ogr_geom.GetGeometryType() == 6:
                    under_ogr_geom = ogr.Geometry(ogr.wkbPolygon)
                elif ogr_geom.GetGeometryType() == 3:
                    under_ogr_geom = ogr.Geometry(ogr.wkbLinearRing)
                elif ogr_geom.GetGeometryType() == 5:
                    under_ogr_geom = ogr.Geometry(ogr.wkbLineString)

                for i, coordinates_list in enumerate(coordinates):
                    if ogr_geom.GetGeometryType() == 2:
                        if i == 0:
                            if ogr_geom.GetGeometryName() == 'LINESTRING':
                                ogr_geom = ogr.Geometry(ogr.wkbLineString)
                            else:
                                ogr_geom = ogr.Geometry(ogr.wkbLinearRing)
                        ogr_geom.AddPoint_2D(coordinates_list[0], coordinates_list[1])

                    else:
                        if i == 0 and ogr_geom.GetGeometryType() == 3:
                            ogr_geom = ogr.Geometry(ogr.wkbPolygon)
                        new_geom = coordinates_loop(coordinates_list, under_ogr_geom)
                        ogr_geom.AddGeometry(new_geom)

        return ogr_geom

    if import_ogr_success:
        geo_type = geometry['type']
        if geo_type.upper() == 'POINT':
            ogr_geom = ogr.Geometry(ogr.wkbPoint)
        elif geo_type.upper() == 'LINESTRING':
            ogr_geom = ogr.Geometry(ogr.wkbLineString)
        elif geo_type.upper() == 'POLYGON':
            ogr_geom = ogr.Geometry(ogr.wkbPolygon)
        elif geo_type.upper() == 'MULTIPOINT':
            ogr_geom = ogr.Geometry(ogr.wkbMultiPoint)
        elif geo_type.upper() == 'MULTILINESTRING':
            ogr_geom = ogr.Geometry(ogr.wkbMultiLineString)
        elif geo_type.upper() == 'MULTIPOLYGON':
            ogr_geom = ogr.Geometry(ogr.wkbMultiPolygon)
        elif geo_type.upper() == 'GEOMETRYCOLLECTION':
            ogr_geom = ogr.Geometry(ogr.wkbGeometryCollection)
        elif geo_type.upper() == 'NO GEOMETRY':
            ogr_geom = ogr.Geometry(ogr.wkbNone)
        elif geo_type.upper() == 'GEOMETRY':
            ogr_geom = ogr.Geometry(ogr.wkbUnknown)
        else:
            ogr_geom = ogr.Geometry(ogr.wkbUnknown)

        if geo_type.upper() != 'NO GEOMETRY':
            if geo_type.upper() != 'GEOMETRYCOLLECTION':
                ogr_geom = coordinates_loop(geometry['coordinates'], ogr_geom)
            else:
                ogr_geom = coordinates_loop(geometry['geometries'], ogr_geom)

        return ogr_geom
    else:
        raise Exception(import_ogr_error)


def geometry_to_wkb(geometry, endian_big=True):
    """
    Transform geoformat geometry (or geojson like geometry) to wkb geometry.
    Optionally you can choose the output endian.

    :param geometry: geoformat geometry (or geojson like geometry)
    :param endian_big: True if output big endian / False if output little endian.
    :return: output bytes geometry in wkb format
    """

    # START geometry_to_wkb
    if endian_big is True:
        endian_b = b'\x00'
    else:
        endian_b = b'\x01'

    geojson_type = geometry['type'].upper()
    multi_geometry = False
    if 'MULTI' in geojson_type:
        multi_geometry = True

    geometry_collection = False
    if geojson_type == 'GEOMETRYCOLLECTION':
        geometry_collection = True

    bytes_geo_type = bytearray(geometry_type_to_wkb_geometry_type[geojson_type])
    if endian_big is False:
        bytes_geo_type.reverse()

    if geometry_collection:
        if geometry['geometries']:
            wkb_coordinates = int_to_4_bytes_integer(integer_value=len(geometry['geometries']),
                                                     integer_endian_big=endian_big)
            for geometry_from_collection in geometry['geometries']:
                wkb_coordinates += geometry_to_wkb(geometry=geometry_from_collection, endian_big=endian_big)
        else:
            wkb_coordinates = b'\x00\x00\x00\x00'
    elif multi_geometry:
        if geometry['coordinates']:
            wkb_coordinates = int_to_4_bytes_integer(integer_value=len(geometry['coordinates']),
                                                     integer_endian_big=endian_big)
            for single_geom in multi_geometry_to_single_geometry(geometry, bbox=False):
                wkb_coordinates += geometry_to_wkb(geometry=single_geom, endian_big=endian_big)
        else:
            wkb_coordinates = b'\x00\x00\x00\x00'
    else:
        if geometry['coordinates']:
            wkb_coordinates = coordinates_list_to_bytes(geometry['coordinates'], coordinates_big_endian=endian_big)
        else:
            if geojson_type == 'POINT':
                if endian_big:
                    wkb_coordinates = b'\x7f\xf8\x00\x00\x00\x00\x00\x00\x7f\xf8\x00\x00\x00\x00\x00\x00'
                else:
                    wkb_coordinates = b'\x00\x00\x00\x00\x00\x00\xf8\x7f\x00\x00\x00\x00\x00\x00\xf8\x7f'
            else:
                wkb_coordinates = b'\x00\x00\x00\x00'

    wkb_bytearray = endian_b + bytes_geo_type + wkb_coordinates

    return wkb_bytearray


def wkb_to_geometry(wkb_geometry, bbox=True):
    """
    Transform wkb bytes geometry to geoformat (geojson like) geometry

    :param wkb_geometry: input wkb bytes
    :param bbox: boolean that indicate if we want or not the bbox information in geometry
    :return: geoformat geometry (geojson like)
    """

    def define_geometry_lenght_in_bytesarray(geometry_b, geometry_start_idx):
        """
        Return the geometry length (in number of bytes).

        :param geometry_b: geometry in wkb.
        :param geometry_start_idx: start index of geometrie in geometry_b.
        :return: length in integer.
        """
        # determine endian
        endian_b = geometry_b[geometry_start_idx]
        part_big_endian = True
        if endian_b == 1:
            part_big_endian = False
        geometry_start_idx += 1
        # determine geo type
        geometry_end_idx = geometry_start_idx + 4
        _geo_type_b = geometry_b[geometry_start_idx:geometry_end_idx]
        if part_big_endian is False:
            _geo_type_b = bytearray(_geo_type_b)
            _geo_type_b.reverse()
            _geo_type_b = bytes(_geo_type_b)
        _geo_type = big_endian_wkb_geometry_type_to_geometry_type[_geo_type_b]
        # deduce double_dimension
        _dimension = geometry_type_to_dimension[_geo_type]
        geometry_start_idx = geometry_end_idx
        geometry_end_idx = geometry_end_idx + 4
        if 'Multi' in _geo_type:
            _geometry_idx_length = 9
            nb_part_b = geometry_b[geometry_start_idx:geometry_end_idx]
            nb_part = integer_4_bytes_to_int(integer_4_bytes=nb_part_b, integer_endian_big=part_big_endian)
            geometry_start_idx = geometry_end_idx
            for i_part in range(nb_part):
                length_part = define_geometry_lenght_in_bytesarray(geometry_b=geometry_b,
                                                                   geometry_start_idx=geometry_start_idx)
                _geometry_idx_length += length_part
                geometry_start_idx += length_part
        elif _geo_type == 'Point':
            _geometry_idx_length = 5 + _dimension * 8
        else:
            # determine nb coordinates or rings
            _nb_coordinates_or_rings_b = geometry_b[geometry_start_idx:geometry_end_idx]
            _nb_coordinates_or_rings = integer_4_bytes_to_int(integer_4_bytes=_nb_coordinates_or_rings_b,
                                                              integer_endian_big=part_big_endian)
            if 'LineString' in _geo_type:
                # deduce length of linestring part
                _geometry_idx_length = 9 + _dimension * 8 * _nb_coordinates_or_rings
            elif 'Polygon' in _geo_type:
                _geometry_idx_length = 9  # endian + _geo_type + nb_ring
                geometry_start_idx = geometry_end_idx
                # loop on each ring
                for ring in range(_nb_coordinates_or_rings):
                    # deduce nb coordinates by ring
                    geometry_end_idx = geometry_start_idx + 4
                    _nb_coordinates_b = geometry_b[geometry_start_idx:geometry_end_idx]
                    _nb_coordinates = integer_4_bytes_to_int(integer_4_bytes=_nb_coordinates_b,
                                                             integer_endian_big=part_big_endian)
                    # deduce ring length
                    ring_length = 4 + _dimension * 8 * _nb_coordinates  # nb coordinates + length of coordinates
                    _geometry_idx_length += ring_length
                    geometry_start_idx = geometry_start_idx + ring_length
            else:
                raise Exception('geo type not valid')

        return _geometry_idx_length

    def get_coordinates_list_from_wkb_geometry(wkb_coordinates, nb_coord, wkb_dimension, wkb_endian_big=True):
        """
        Determine coordinates list from wkb coordinates. To do that we need the wkb coordinates, the number of
        coordinates and their number of dimensions.
        Optionally you can choose the output endian.

        :param wkb_coordinates: coordinates in wkb
        :param nb_coord: number of coordinates
        :param wkb_dimension: number of dimensions in wkb coordinates
        :param wkb_endian_big: True if output big endian / False if output little endian.
        :return: coordinates list
        """
        _coordinates_list = [None] * nb_coord
        split_idx_list = range(0, nb_coord * wkb_dimension * 8, wkb_dimension * 8)
        for i_coord, _start_idx in enumerate(split_idx_list):
            _end_idx = _start_idx + 8 * wkb_dimension
            part_coordinates_b = wkb_coordinates[_start_idx:_end_idx]
            float_coordinates_list = list(double_8_bytes_to_float(double_8_bytes=part_coordinates_b,
                                                                  double_big_endian=wkb_endian_big,
                                                                  double_dimension=2))
            _coordinates_list[i_coord] = float_coordinates_list

        return _coordinates_list

    # START wkb_to_geometry
    # get endianess of geometry
    geometry_endian_b = wkb_geometry[0]
    if geometry_endian_b == 0:
        endian_big = True
    elif geometry_endian_b == 1:
        endian_big = False
    else:
        raise ValueError('wkb format must begin by "\x00" for BIG or "\x01" for LITTLE endian')
    # get geotype of geometry
    geo_type_b = wkb_geometry[1:5]
    if endian_big is False:
        geo_type_b = bytearray(geo_type_b)
        geo_type_b.reverse()
        geo_type_b = bytes(geo_type_b)
    geo_type = big_endian_wkb_geometry_type_to_geometry_type[geo_type_b]

    # init output geometry
    return_geometry = {'type': geo_type}
    # init idx
    start_idx = 5
    end_idx = start_idx + 4

    if geo_type == 'GeometryCollection':
        # get number of geometries
        nb_geometries_b = wkb_geometry[start_idx:end_idx]
        nb_geometries = integer_4_bytes_to_int(integer_4_bytes=nb_geometries_b, integer_endian_big=endian_big)
        start_idx = end_idx
        geometries_list = [None] * nb_geometries
        if bbox is True:
            geometry_collection_bbox = ()
        for idx_geometry in range(nb_geometries):
            # get length of geometry in collection (in bytes)
            geometry_in_collection_length = define_geometry_lenght_in_bytesarray(geometry_b=wkb_geometry,
                                                                                 geometry_start_idx=start_idx)
            end_idx = start_idx + geometry_in_collection_length
            # create geometry in collection
            geometry_in_collection_b = wkb_geometry[start_idx:end_idx]
            geometry_in_collection = wkb_to_geometry(geometry_in_collection_b, bbox=bbox)
            # add geometry in geometries_list
            geometries_list[idx_geometry] = geometry_in_collection
            # reset start idx
            start_idx = end_idx
            # add bbox if option is True
            if bbox is True:
                if 'bbox' in geometry_in_collection:
                    geometry_bbox = geometry_in_collection['bbox']
                    if not geometry_collection_bbox:
                        geometry_collection_bbox = geometry_bbox
                    else:
                        geometry_collection_bbox = bbox_union(geometry_bbox, geometry_collection_bbox)

        # add geometries to list
        return_geometry['geometries'] = geometries_list
        # if bbox is True
        if bbox is True and geometry_collection_bbox:
            return_geometry['bbox'] = geometry_collection_bbox

    else:
        dimension = geometry_type_to_dimension[geo_type]
        if 'Multi' in geo_type:
            # idx for nb geometries (4 bytes)
            end_idx = start_idx + 4
            nb_geometries_b = wkb_geometry[start_idx:end_idx]
            nb_geometries = integer_4_bytes_to_int(integer_4_bytes=nb_geometries_b, integer_endian_big=endian_big)
            start_idx = end_idx
            coordinates_list = [None] * nb_geometries
            for i_geometry in range(nb_geometries):
                # get length of geometry
                geometry_idx_length = define_geometry_lenght_in_bytesarray(geometry_b=wkb_geometry,
                                                                           geometry_start_idx=start_idx)
                end_idx = start_idx + geometry_idx_length
                slice_geom_b = wkb_geometry[start_idx:end_idx]
                slice_geom = wkb_to_geometry(slice_geom_b, bbox=False)
                coordinates_list[i_geometry] = slice_geom['coordinates']
                # reset start_idx for next geometry
                start_idx = end_idx
        elif geo_type == 'Point':
            x_y_b = wkb_geometry[start_idx:]
            # check if empty geometry
            if (endian_big and x_y_b == b'\x7f\xf8\x00\x00\x00\x00\x00\x00\x7f\xf8\x00\x00\x00\x00\x00\x00') or (
                    not endian_big and x_y_b == b'\x00\x00\x00\x00\x00\x00\xf8\x7f\x00\x00\x00\x00\x00\x00\xf8\x7f'):
                coordinates_list = []
            # if not empty geometry get coordinates
            else:
                coordinates = double_8_bytes_to_float(double_8_bytes=x_y_b,
                                                      double_big_endian=endian_big,
                                                      double_dimension=dimension)
                coordinates_list = list(coordinates)
        else:
            # get number of coordinates
            nb_coordinates_or_rings_b = wkb_geometry[start_idx:end_idx]
            nb_coordinates_or_rings = integer_4_bytes_to_int(integer_4_bytes=nb_coordinates_or_rings_b,
                                                             integer_endian_big=endian_big)
            if geo_type == 'LineString':
                len_bytes = nb_coordinates_or_rings * 8 * dimension
                start_idx = end_idx
                end_idx = start_idx + len_bytes
                coordinates_in_wkb = wkb_geometry[start_idx:end_idx]
                # get coordinates
                coordinates_list = get_coordinates_list_from_wkb_geometry(nb_coord=nb_coordinates_or_rings,
                                                                          wkb_coordinates=coordinates_in_wkb,
                                                                          wkb_dimension=dimension,
                                                                          wkb_endian_big=endian_big)
            elif geo_type == 'Polygon':
                # get coordinates in each rings
                ring_list = [None] * nb_coordinates_or_rings
                for i_ring in range(nb_coordinates_or_rings):
                    start_idx = end_idx
                    end_idx = start_idx + 4
                    nb_coordinates_b = wkb_geometry[start_idx:end_idx]
                    # get number of coordinates
                    nb_coordinates = integer_4_bytes_to_int(integer_4_bytes=nb_coordinates_b,
                                                            integer_endian_big=endian_big)
                    len_bytes = nb_coordinates * 8 * dimension
                    start_idx = end_idx
                    end_idx = start_idx + len_bytes
                    coordinates_in_wkb = wkb_geometry[start_idx:end_idx]
                    # get coordinates
                    coordinates_list = get_coordinates_list_from_wkb_geometry(nb_coord=nb_coordinates,
                                                                              wkb_coordinates=coordinates_in_wkb,
                                                                              wkb_dimension=dimension,
                                                                              wkb_endian_big=endian_big)
                    # write coordinates in rings list
                    ring_list[i_ring] = coordinates_list

                coordinates_list = ring_list
            else:
                raise Exception('error on geometry type')

        # add coordinates to geom
        return_geometry['coordinates'] = coordinates_list

        # add bbox if true and not empty geometry
        if bbox is True and coordinates_list:
            return_geometry['bbox'] = coordinates_to_bbox(coordinates_list)

    return return_geometry


def geometry_to_wkt(geometry):
    """
    Convert geoformat geometry to WKT geometry.

    :param geometry: input geometry
    :return: wkt geometry
    """

    def coordinates_list_to_wkt_coordinates(coordinates_list_or_tuple, coordinates_dimension=2):
        """
        Loop over coordinates and return a wkt string of this coordinates

        :param coordinates_list_or_tuple: original list of coordinates
        :param coordinates_dimension: number of dimension for each coordinate vertex
        :return: wkt format coordinates
        """

        if coordinates_list_or_tuple:
            _coordinates_pattern = ' '.join(['{}'] * coordinates_dimension)

            if isinstance(coordinates_list_or_tuple[0], (float, int)):
                _coordinates_txt = _coordinates_pattern.format(*coordinates_list_or_tuple)
            else:  # then it's iterable data
                _coordinates_txt = ''
                for _coordinates in coordinates_list_or_tuple:
                    _coordinates_txt += coordinates_list_to_wkt_coordinates(coordinates_list_or_tuple=_coordinates,
                                                                            coordinates_dimension=coordinates_dimension) + ','
                _coordinates_txt = '(' + _coordinates_txt[:-1] + ')'
        else:
            _coordinates_txt = "EMPTY"

        return _coordinates_txt

    # START geometry_to_wkt
    geo_type_upper = geometry['type'].upper()

    if geo_type_upper == 'GEOMETRYCOLLECTION':
        if geometry['geometries']:
            coordinates_txt = ''
            for _geometry in geometry['geometries']:
                coordinates_txt += geometry_to_wkt(geometry=_geometry) + ','
            coordinates_txt = '(' + coordinates_txt[:-1] + ')'
        else:
            coordinates_txt = 'EMPTY'
    else:
        # deduce coordinates_dimension
        dimension = geometry_type_to_dimension[geometry['type']]
        coordinates_txt = coordinates_list_to_wkt_coordinates(coordinates_list_or_tuple=geometry['coordinates'],
                                                              coordinates_dimension=dimension)
        if geo_type_upper == 'POINT' and coordinates_txt != 'EMPTY':
            coordinates_txt = '(' + coordinates_txt + ')'

    return geo_type_upper + ' ' + coordinates_txt


def wkt_to_geometry(wkt_geometry, bbox=True):
    """
    Make translation between wkt geometry and geoformat geometry.

    :param wkt_geometry: WKT geometry that we want to convert to geoformat geometry
    :param bbox: boolean that indicate if we want or not the bbox information in geometry
    :return: geoformat geometry
    """

    def get_wkt_geometries_in_collection(wkt_geometries):
        """
        return a list with wkt geometrie from geometrycollection

        :param wkt_geometries:
        :return: list with wkt geometries in it
        """
        # make a list of allowed geometry : ordered count here (when we will use find())
        geometry_type_allowed_in_collection_ordered = ["POINT", "LINESTRING", "POLYGON", "MULTIPOINT",
                                                       "MULTILINESTRING", "MULTIPOLYGON"]
        # create reversed geometry type
        reversed_geotype_list = [geotype[::-1] for geotype in geometry_type_allowed_in_collection_ordered]

        # delete starting and finishing parenthesis
        geometry_in_collection = wkt_geometries[1:-1]
        # reverse coordinates string
        coordinates_wkt_reversed = geometry_in_collection[::-1]

        geometry_in_collection_list = []
        # loop over coordinates_wkt_reversed as long their are data
        while len(coordinates_wkt_reversed) > 0:
            min_find_idx = len(coordinates_wkt_reversed)
            # loop on each reversed geotype in reversed coordinates_wkt_reversed
            for i_geotype, reversed_geotype in enumerate(reversed_geotype_list):
                # determine geotype position in coordinates_wkt_reversed
                find_idx = coordinates_wkt_reversed.find(reversed_geotype)
                # if index is superior than -1 (no match) and inferior to min_find_idx
                if -1 < find_idx <= min_find_idx:
                    min_find_idx = find_idx
                    min_geotype_reversed = reversed_geotype

            # resize index for entire geometry
            geom_idx = min_find_idx + len(min_geotype_reversed)
            # extract geometry
            reversed_geom = coordinates_wkt_reversed[:geom_idx]
            # reverse geom
            geom = reversed_geom[::-1]
            # store geom in geometry_in_collection_list
            geometry_in_collection_list.append(geom)

            # resize coordinates_wkt_reversed
            coordinates_wkt_reversed = coordinates_wkt_reversed[geom_idx + 1:]

        # reverse list to have original geometries order
        geometry_in_collection_list.reverse()

        return geometry_in_collection_list

    def float_or_int_coordinate_str(coordinate_str):
        """
        Make translation between string coordinates to integer to float coordinates.
        EMPTY geometry doesn't exist in geoformat => this function return None value.

        :param coordinate_str: coordinates in string
        :return: corrdinates in int (if no decimal number) or float (if decimal number)
        """
        if '.' in coordinate_str:
            coordinate = float(coordinate_str)
        else:
            coordinate = int(coordinate_str)

        return coordinate

    if '(' in wkt_geometry:
        geo_type_txt = wkt_geometry.split('(')[0].strip().upper()
        output_geometry = {'type': upper_case_to_geometry_type[geo_type_txt]}
        coordinates_wkt = wkt_geometry.split(wkt_geometry.split('(')[0].lstrip())[1]
        if geo_type_txt == 'GEOMETRYCOLLECTION':
            # retrieve in a list geometries stored in coordinates_wkt
            wkt_geometries_list = get_wkt_geometries_in_collection(coordinates_wkt)
            # convert wkt_geometry to geoformat geometries
            output_geometry["geometries"] = [None] * len(wkt_geometries_list)
            if bbox is True:
                geometry_collection_bbox = ()
            for i_geometry, wkt_geometry in enumerate(wkt_geometries_list):
                geoformat_geometry = wkt_to_geometry(wkt_geometry, bbox=bbox)
                # add bbox parameters for geometry and geometry_collection
                if bbox is True:
                    # if geometry is not
                    if 'bbox' in geoformat_geometry:
                        geometry_bbox = geoformat_geometry['bbox']
                        geoformat_geometry['bbox'] = geometry_bbox
                        if not geometry_collection_bbox:
                            geometry_collection_bbox = geometry_bbox
                        else:
                            geometry_collection_bbox = bbox_union(geometry_bbox, geometry_collection_bbox)
                # add geometry in output_geometry
                output_geometry['geometries'][i_geometry] = geoformat_geometry

            # write bbox for geometries
            if bbox is True:
                if geometry_collection_bbox:
                    output_geometry['bbox'] = geometry_collection_bbox
        else:
            # Now we extract coordinates from coordinates_wkt.
            # all geometrie will have same number of parenthesis in coordinates_wkt
            # like this all coordinates will have same depth when we will split str (coordinates_wkt) to extract coordinates
            first_deep = {'POINT', 'LINESTRING', 'POLYGON', 'MULTIPOINT', 'MULTILINESTRING'}
            second_deep = {'POINT', 'LINESTRING', 'MULTIPOINT'}

            # add parentesis
            if geo_type_txt in first_deep:
                coordinates_wkt = '(' + coordinates_wkt + ')'
                if geo_type_txt in second_deep:
                    coordinates_wkt = '(' + coordinates_wkt + ')'

            # for multi point without extra parenthesis
            multipoint_with_extra_parenthesis = False
            if geo_type_txt == 'MULTIPOINT' and set(coordinates_wkt[-4:]) == set(')'):
                coordinates_wkt = coordinates_wkt[1:-1]
                multipoint_with_extra_parenthesis = True

            # split string  coordinates_wkt and iterates through the depth of the coordinates
            coordinates_wkt = coordinates_wkt.split('(((')[1].split(')))')[0]
            coordinates_part = coordinates_wkt.split('((')
            geometry_coordinates_list = [None] * len(coordinates_part)
            for i_part, part in enumerate(coordinates_part):
                format_part = part.split('))')[0]
                coordinates_ring = format_part.split('(')
                geometry_coordinates_list[i_part] = [None] * len(coordinates_ring)
                for i_ring, ring in enumerate(coordinates_ring):
                    ring_format = ring.split(')')[0].replace(', ', ',')
                    coordinates_list = ring_format.split(',')
                    geometry_coordinates_list[i_part][i_ring] = [None] * len(coordinates_list)
                    for i_coord, coordinates in enumerate(coordinates_list):
                        # extract x y coordinates
                        coordinates_x_y = coordinates.split(',')[0]
                        coordinates_x_y = coordinates_x_y.split(' ')
                        # convert coordinates to int or float
                        x = float_or_int_coordinate_str(coordinates_x_y[0])
                        y = float_or_int_coordinate_str(coordinates_x_y[1])
                        coordinates_result = [x, y]
                        # store coordinates in geometry_coordinates_list
                        if multipoint_with_extra_parenthesis:
                            geometry_coordinates_list[i_part][i_ring] = coordinates_result
                        else:
                            geometry_coordinates_list[i_part][i_ring][i_coord] = coordinates_result
                        # calculate bbox
                        if bbox is True:
                            # init bbox
                            if i_part == 0 and i_ring == 0 and i_coord == 0:
                                geometry_bbox = [x, y, x, y]
                            # udpate bbox
                            else:
                                if x < geometry_bbox[0]:
                                    geometry_bbox[0] = x
                                if y < geometry_bbox[1]:
                                    geometry_bbox[1] = y
                                if x > geometry_bbox[2]:
                                    geometry_bbox[2] = x
                                if y > geometry_bbox[3]:
                                    geometry_bbox[3] = y

            # recreate deepest of coordinates list
            if geo_type_txt in first_deep:
                geometry_coordinates_list = geometry_coordinates_list[0]
                if geo_type_txt in second_deep and multipoint_with_extra_parenthesis is False:
                    geometry_coordinates_list = geometry_coordinates_list[0]
                    if geo_type_txt == 'POINT':
                        geometry_coordinates_list = geometry_coordinates_list[0]
            # save coordinate list in geomet
            output_geometry['coordinates'] = geometry_coordinates_list

            # add bbox
            if bbox is True:
                output_geometry['bbox'] = tuple(geometry_bbox)
    else:
        if 'EMPTY' in wkt_geometry:
            geo_type_txt = wkt_geometry.split('EMPTY')[0].strip().upper()
            if geo_type_txt == 'GEOMETRYCOLLECTION':
                output_geometry = {'type': upper_case_to_geometry_type[geo_type_txt], 'geometries': []}
            else:
                output_geometry = {'type': upper_case_to_geometry_type[geo_type_txt], 'coordinates': []}
        else:
            raise Exception('not WKT linestring')

    return output_geometry


def force_rhr(polygon_geometry):
    """
    Forces the orientation of the vertices in a polygon to follow a Right-Hand-Rule, in which the area that is bounded
    by the polygon is to the right of the boundary. In particular, the exterior ring is orientated in a clockwise
    direction and the interior rings in a counter-clockwise direction.

    Equivalent to Postgis : st_forcerhr

    :param polygon_geometry: polygon or multipolygon geometry
    :return: right hand rule polygon or multipolygon
    """
    # filter on geometry type (ONLY POLYGON and MULTIPOLYGON are allowed)
    if 'POLYGON' in polygon_geometry["type"].upper():
        # create output coordinates list (this list will have the same
        # deepest for polygon and multipolygon : we will correct this after).
        if 'MULTI' in polygon_geometry["type"].upper():
            output_coordinates = [None] * len(polygon_geometry['coordinates'])
        else:
            output_coordinates = [None]

        # loop on each part of polygon
        for i_geom, geometry in enumerate(multi_geometry_to_single_geometry(polygon_geometry, bbox=False)):
            rhr_geometry = force_rhr_polygon_coordinates(coordinates=geometry["coordinates"])
            output_coordinates[i_geom] = rhr_geometry

        # if input geometry is polygon we restore the deepest of coordinates list
        if polygon_geometry['type'].upper() == 'POLYGON':
            output_coordinates = output_coordinates[0]

        # store result in output geometry
        output_geometry = {"type": polygon_geometry["type"], "coordinates": output_coordinates}

        # if bbox in geometry
        if "bbox" in polygon_geometry:
            output_geometry["bbox"] = polygon_geometry["bbox"]
    else:
        raise Exception(geometry_must_be_polygon_or_multipolygon)

    return output_geometry


def geometry_to_bbox(geometry):
    """
    Return bbox of input geometry
    """

    geometry_collection = geometry_to_geometry_collection(geometry, bbox=False)
    geometry_bbox = None
    if geometry_collection['geometries']:
        for geometry in geometry_collection['geometries']:
            bbox = coordinates_to_bbox(geometry['coordinates'])
            if geometry_bbox:
                geometry_bbox = bbox_union(bbox, geometry_bbox)
            else:
                geometry_bbox = bbox
    else:
        geometry_bbox = ()

    return geometry_bbox


def reproject_geometry(geometry, in_crs, out_crs, precision=None):
    """
    Reproject a geometry with an input coordinate reference system to an output coordinate system

    :param geometry: input geometry
    :param in_crs: input coordinate reference system
    :param out_crs: output coordinate system
    :param precision:
    :return: output geometry
    """

    input_geometry_type = geometry["type"]
    # transform geometry in geometry_collection collection
    geometry_collection = geometry_to_geometry_collection(geometry=geometry, bbox=False)
    # loop on each geometries
    for geometry_in_collection in geometry_collection['geometries']:
        reproj_coordinates = format_coordinates(
            coordinates_list_tuple=geometry_in_collection['coordinates'],
            precision=precision,
            in_crs=in_crs,
            out_crs=out_crs
        )
        geometry_in_collection['coordinates'] = reproj_coordinates
    # if input geometry is not GeometryCollection we return geometry's original type
    if input_geometry_type != 'GeometryCollection':
        output_geometry = geometry_collection['geometries'][0]
    else:
        output_geometry = geometry_collection

    return output_geometry




from geoformat.geoprocessing.generalization.ramer_douglas_peucker import ramer_douglas_peucker
from geoformat.geoprocessing.generalization.visvalingam_whyatt import visvalingam_whyatt
from geoformat.conversion.geometry_conversion import geometry_to_geometry_collection, multi_geometry_to_single_geometry


def simplify(geometry, tolerance, algo='RDP'):
    """
    Generalize input geometry to a given tolerance with a specified generalization algorithm.

    algorithm list :
        RDP : Ramer Douglas Peucker
        VW : Visvalingam Whyatt

    Compare Rameer Douglas Peucker and Visvalingam Whyatt :
        http://bl.ocks.org/msbarry/9152218
        https://martinfleischmann.net/line-simplification-algorithms/

    To visually test the algorithms one can use: https://mapshaper.org/

    :param geometry: geometry to generalize.
    :param tolerance: tolerance threshold for generalization.
    :param algo: type of algorithm used to make the generalisation (two values allowed RDP or VW).
    :return: generalized geometry
    """

    # init algo
    if algo == 'RDP':
        simplify_algorithm = ramer_douglas_peucker
    elif algo == 'VW':
        simplify_algorithm = visvalingam_whyatt
    else:
        raise Exception(
            "You have to choose between two values : 'RDP' (Ramer Douglas Peucker) or 'VW' (Visvalingam Whyatt")

    input_geometry_type = geometry["type"]
    # transform geometry in geometry_collection collection
    geometry_collection = geometry_to_geometry_collection(geometry=geometry, bbox=False)
    result_geometry_collection = {"type": 'GeometryCollection', "geometries": []}
    # loop on each geometries
    for geometry_in_collection in geometry_collection['geometries']:
        in_collection_geometry_coordinates = []
        # if geometry_collection is a multi geometry_collection loop on each single geometry_collection
        for single_geometry in multi_geometry_to_single_geometry(geometry_in_collection, bbox=False):
            # if single geometry_collection is LineString or Polygon we can simplify it
            if single_geometry['type'] in {"LineString", "Polygon"}:
                # create a basic coordinates list to launch ramer douglas peucker in it
                coordinates_list = single_geometry['coordinates']
                if single_geometry['type'] == 'LineString':
                    coordinates_list = [single_geometry['coordinates']]
                # loop on each coordinates
                result_coordinates_list = []
                for coordinates_list in coordinates_list:
                    result_coordinates = (simplify_algorithm(coordinate_list=coordinates_list, tolerance=tolerance))
                    result_coordinates_list.append(result_coordinates)

                if single_geometry['type'] == 'LineString':
                    result_coordinates_list = result_coordinates_list[0]

            # if single_geometry is point (we do nothing)
            else:
                result_coordinates_list = single_geometry['coordinates']

            in_collection_geometry_coordinates.append(result_coordinates_list)

        if 'Multi' not in geometry_in_collection['type']:
            in_collection_geometry_coordinates = in_collection_geometry_coordinates[0]

        result_geometry_in_collection = {
            "type": str(geometry_in_collection['type']),
            "coordinates": in_collection_geometry_coordinates
        }
        result_geometry_collection['geometries'].append(result_geometry_in_collection)

    # if input geometry is not GeometryCollection we return geometry's original type
    if input_geometry_type != 'GeometryCollection':
        output_geometry = result_geometry_collection['geometries'][0]
    else:
        output_geometry = result_geometry_collection

    return output_geometry





from geoformat.geoprocessing.connectors.operations import segment_to_bbox
from geoformat.geoprocessing.geoparameters.lines import (
    line_parameters,
    perpendicular_line_parameters_at_point,
    crossing_point_from_lines_parameters
)


def euclidean_distance(point_a, point_b):
    """
    retunn Euclidean distance between two points

    :param point_a:
    :param point_b:
    :return: Euclidean distance between two points
    """
    (x_a, y_a) = point_a
    (x_b, y_b) = point_b
    return ((x_b - x_a) ** 2 + (y_b - y_a) ** 2) ** 0.5


def manhattan_distance(point_a, point_b):
    """
    return Manhattan distance between to points

    :param point_a:
    :param point_b:
    :return: Manhattan distance between to points
    """
    x1, y1 = point_a
    x2, y2 = point_b
    return float(abs(x1 - x2) + abs(y1 - y2))


def euclidean_distance_point_vs_segment(point, segment):
    """
    Calculate distance between a point and a segment

    adapted from : https://stackoverflow.com/questions/849211/shortest-distance-between-a-point-and-a-line-segment

    Much more faster than point_vs_segment_distance

    :param point: vertex coordinate
    :param segment: segment coordinates
    :return: distance between point and segment
    """

    ((x1, y1), (x2, y2)) = segment
    (x3, y3) = point

    px = x2-x1
    py = y2-y1

    something = px*px + py*py

    # if segment is actually a point
    if something == 0:
        dist = euclidean_distance(point_a=point, point_b=segment[0])
    else:
        u = ((x3 - x1) * px + (y3 - y1) * py) / float(something)

        if u > 1:
            u = 1
        elif u < 0:
            u = 0

        x = x1 + u * px
        y = y1 + u * py

        dx = x - x3
        dy = y - y3

        # Note: If the actual distance does not matter,
        # if you only want to compare what this function
        # returns to other results of this function, you
        # can just return the squared distance instead

        dist = (dx*dx + dy*dy) ** 0.5

    return dist


def point_vs_segment_distance(point, segment, distance_function=euclidean_distance):
    """
    Calculate distance between a point and a segment, distance can be defined with distance function parameter (default :
    euclidean)

    :param point:
    :param segment:
    :param distance_function:
    :return:
    """
    # get line parameter to segment
    segment_line_parameters = line_parameters(segment)
    # get perpendicular parameters to line parameters above
    perpendicular_line_parameters = perpendicular_line_parameters_at_point(line_parameters=segment_line_parameters,
                                                                           point=point)
    # get intersection
    intersection_point = crossing_point_from_lines_parameters(line_parameter_a=segment_line_parameters,
                                                              line_parameter_b=perpendicular_line_parameters)
    # get segment bbox
    segment_bbox = segment_to_bbox(segment)
    # if point is in bbox (we use bbox for floating point precision consideration)
    if segment_bbox[0] <= intersection_point[0] <= segment_bbox[2] and segment_bbox[1] <= intersection_point[1] <= segment_bbox[3]:
        return distance_function(point, intersection_point)
    # else we have to calculate distance between start and end point to segment (we return minimal distance)
    else:
        start_point = segment[0]
        end_point = segment[1]
        return min(distance_function(point, start_point), distance_function(point, end_point))



from geoformat.geoprocessing.geoparameters.lines import (
    line_parameters,
    point_at_distance_with_line_parameters,
    perpendicular_line_parameters_at_point
)
from geoformat.geoprocessing.measure.mesure_distance import (
    euclidean_distance
)

from geoformat.conversion.geometry_conversion import geometry_to_multi_geometry


def point_at_a_distance_on_segment(segment, step_distance, offset_distance=None):
    """
    Iterator send point coordinates on a segment at a given step distance.
    optional: add an offset distance (perpendicular to line_parameters value)

    :param segment:
    :param step_distance:
    :param offset_distance:
    :return:
    """
    # initialize
    native_step_distance = step_distance

    # split segment to point
    start_point, end_point = segment
    # get segment line parameter
    segment_line_parameter = line_parameters(segment)

    # recompute start point and end point and line parameters if offset is required
    if offset_distance:
        start_point = point_at_distance_with_line_parameters(
            start_point,
            segment_line_parameter,
            0,
            offset_distance=offset_distance
        )
        end_point = point_at_distance_with_line_parameters(
            end_point,
            segment_line_parameter,
            0,
            offset_distance=offset_distance
        )
        segment_line_parameter = line_parameters((start_point, end_point))

    # segment_line_parameter slope determine line orientation (from left to the right)
    # we need to to determine the direction on this slope between start_point and end_point
    point_at_a_distance = point_at_distance_with_line_parameters(
        start_point=end_point,
        line_parameters=segment_line_parameter,
        distance=step_distance,
    )

    length_between_start_point_and_point_at_distance = euclidean_distance(
        point_a=start_point,
        point_b=point_at_a_distance
    )
    segment_length = euclidean_distance(
        point_a=start_point,
        point_b=end_point
    )

    if length_between_start_point_and_point_at_distance < segment_length:
        step_distance = -step_distance

    while segment_length >= native_step_distance:
        # compute new start point at a distance
        start_point = point_at_distance_with_line_parameters(
            start_point=start_point,
            line_parameters=segment_line_parameter,
            distance=step_distance
        )
        yield start_point
        # recompute segment length with new start_point
        segment_length = euclidean_distance(
            point_a=start_point,
            point_b=end_point
        )


def points_on_linestring_distance(linestring, step_distance, offset_distance=None):
    """
    Return a point geometry that is a point on a given linestring at a given step_distance
    optional: add an offset distance (perpendicular to line_parameters value)

    :param linestring: LineString or MultiLineString geometry
    :param step_distance: distance between each steps
    :param offset_distance: if you want you can add an offset to final on point value
    :return: Point geometry
    """

    def points_on_linestring_part(coordinates, step_distance, offset_distance=None):

        # note about remaining_distance
        # remaining_distance is the distance that remain after a new vertex (because when there is a new vertex we have
        # to recompute the step_distance remaining

        # loop on each coordinate
        for i_point, point in enumerate(coordinates):
            if i_point == 0:
                previous_point = point
                # init remaining distance
                remaining_distance = 0
            else:
                remaining_step_distance = step_distance - remaining_distance
                segment = (previous_point, point)

                # yield first point
                if i_point == 1:
                    first_point_geometry = {'type': 'Point'}
                    if offset_distance:
                        line_parameter = line_parameters(segment)
                        perp_parameter = perpendicular_line_parameters_at_point(line_parameter, previous_point)
                        first_point = point_at_distance_with_line_parameters(previous_point, perp_parameter,
                                                                             distance=offset_distance)
                        first_point_geometry['coordinates'] = list(first_point)
                    else:
                        first_point_geometry['coordinates'] = list(previous_point)

                    yield first_point_geometry

                # for just one iteration
                for new_point in point_at_a_distance_on_segment(segment, remaining_step_distance, offset_distance=None):
                    remaining_distance = 0
                    # reinit values
                    previous_point = new_point
                    segment = (previous_point, point)
                    # here we cannot use offset_distance directly in point_at_a_distance_on_segment used above because
                    # we have to keep initial segment direction. Then we recompute offset new_point.
                    if offset_distance:
                        line_parameter = line_parameters(segment)
                        perp_parameter = perpendicular_line_parameters_at_point(line_parameter, new_point)
                        new_point = point_at_distance_with_line_parameters(new_point, perp_parameter,
                                                                           distance=offset_distance)
                    yield {'type': 'Point', 'coordinates': list(new_point)}
                    break  # just one iteration

                # pass_on_loop check if we iterate on loop below
                pass_on_loop = False
                for new_point in point_at_a_distance_on_segment(segment, step_distance, offset_distance=None):
                    remaining_distance = euclidean_distance(new_point, point)
                    pass_on_loop = True
                    # here we cannot use offset_distance directly in point_at_a_distance_on_segment used above because
                    # we have to calculate the remain distance on non offseted point before. Then we recompute offset
                    # new_point
                    if offset_distance:
                        line_parameter = line_parameters(segment)
                        perp_parameter = perpendicular_line_parameters_at_point(line_parameter, new_point)
                        new_point = point_at_distance_with_line_parameters(new_point, perp_parameter,
                                                                           distance=offset_distance)
                    yield {'type': 'Point', 'coordinates': list(new_point)}

                # if no iteration en loop above we recalculate remaining distance for next point on coordinates
                if not pass_on_loop:
                    remaining_distance += euclidean_distance(point, previous_point)

                previous_point = point

    # force linestring to multilinestring
    multilinestring = geometry_to_multi_geometry(geometry=linestring, bbox=False)

    for linestring_part in multilinestring['coordinates']:
        for point in points_on_linestring_part(linestring_part, step_distance, offset_distance=offset_distance):
            yield point




from geoformat.geoprocessing.measure.mesure_area import shoelace_formula


def ccw_or_cw_boundary(boundary_coordinates):
    """
    The aim of this function is to determinate if boundary[1] of polygon is counter clock wise [CCW] or clock wise [CW].
    To do that we use : Shoelace Formula to determine the area of polygon.

    [1] = here we talk about boundary not entire polygon. We voluntary separate, boundary and hole in polygon, because
    logicaly a boundary can have a sens and the hole can have an other sens. So if you want to know polygon sens you
    have to test separately boundary and hole.

        Input:
            ring_coordinates(list/tuple) :  boundary coordinates

        Output:
            sens: CCW: counter clock wise OR CW clock wise
    """

    if shoelace_formula(boundary_coordinates, False) < 0:
        return 'CCW'
    else:
        return 'CW'




def bbox_expand(bbox, expand):
    """
    Realize ratio enlargement or reduction of a given bbox

    :param bbox: input bbox
    :param expand:  size of enlargement (positive) or reduction (negative)
    :return: enlarge or reduce bbox
    """
    if bbox:
        (x_min, y_min, x_max, y_max) = bbox
    else:
        x_min, y_min, x_max, y_max = (0, 0, 0, 0)

    x_min = x_min - expand
    y_min = y_min - expand
    x_max = x_max + expand
    y_max = y_max + expand

    return x_min, y_min, x_max, y_max

def bbox_union(bbox_a, bbox_b):
    """
    realize union between two given bbox

    :param bbox_a: first bbox
    :param bbox_b: second bbox
    :return: unioned bbox
    """
    if bbox_a and bbox_b:
        (x_min_a, y_min_a, x_max_a, y_max_a) = bbox_a
        (x_min_b, y_min_b, x_max_b, y_max_b) = bbox_b

        bbox = min(x_min_a, x_min_b), min(y_min_a, y_min_b), max(x_max_a, x_max_b), max(y_max_a, y_max_b)
    elif bbox_a and not bbox_b:
        bbox = bbox_a
    elif bbox_b and not bbox_a:
        bbox = bbox_b
    elif not bbox_a and not bbox_b:
        bbox = ()

    return bbox


def point_bbox_position(point, bbox):
    """
    Return point position and sector (NW, N, NE, W, E, SW, S, SE) in regard to given bbox.

    Diagram showing sectors's position around a bbox :

       NW  |   N  |  NE
    -------+------+-------
        W  | bbox |   E
    -------+------+-------
       SW  |   S  |  SE

    3 position possibilities and sectors configuration :
        * Boundary : and side of bbox boundary (N, S, E, W) or corner (NW, NE, SW, SE)
        * Exterior : (NW, N, NE, W, E, SW, S, SE)
        * Interior : None

    """

    (pt_x, pt_y) = point
    (x_min, y_min, x_max, y_max) = bbox

    # North
    if (pt_x > x_min and pt_x < x_max) and (pt_y >= y_max):
        if pt_y == y_max:
            position = ('Boundary', 'N')
        else:
            position = ('Exterior', 'N')
    # South
    elif (pt_x > x_min and pt_x < x_max) and (pt_y <= y_min):
        if pt_y == y_min:
            position = ('Boundary', 'S')
        else:
            position = ('Exterior', 'S')
    # Est
    elif pt_x >= x_max and (pt_y > y_min and pt_y < y_max):
        if pt_x == x_max:
            position = ('Boundary', 'E')
        else:
            position = ('Exterior', 'E')
    # West
    elif pt_x <= x_min and (pt_y > y_min and pt_y < y_max):
        if pt_x == x_min:
            position = ('Boundary', 'W')
        else:
            position = ('Exterior', 'W')
    # North-West
    elif pt_x <= x_min and pt_y >= y_max:
        if pt_x == x_min and pt_y == y_max:
            position = ('Boundary', 'NW')
        else:
            position = ('Exterior', 'NW')
    # North-Est
    elif pt_x >= x_max and pt_y >= y_max:
        if pt_x == x_max and pt_y == y_max:
            position = ('Boundary', 'NE')
        else:
            position = ('Exterior', 'NE')
    # South-Est
    elif pt_x >= x_max and pt_y <= y_min:
        if pt_x == x_max and pt_y == y_min:
            position = ('Boundary', 'SE')
        else:
            position = ('Exterior', 'SE')
    # South-West
    elif pt_x <= x_min and pt_y <= y_min:
        if pt_x == x_min and pt_y == y_min:
            position = ('Boundary', 'SW')
        else:
            position = ('Exterior', 'SW')
    # Point in bbox
    else:
        position = ('Interior', None)

    return position





def _get_line(fields_value_list, max_width_value_list, separator='|'):
    """
    Return a string with values contains in fields_value_list separate by seperator and with for column width the
    informations contained in max_width_value_list

    :param fields_value_list: list of values that we want print
    :param max_width_value_list: list of width column for each values that we want print
    :param separator: separator character between each column
    :return: the formated values contain in fields_value_list
    """

    width_field = [(' ' * (width_value + 2))[:1] + str(fields_value_list[i_field]) + (' ' * (
            width_value + 2))[len(str(fields_value_list[i_field])) + 1:] for i_field, width_value in
     enumerate(max_width_value_list)]

    line = separator + separator.join(width_field) + separator

    return line


def _get_header_separator(width_column_list, table_type='RST'):
    """
    Return a string header of data table

    :param width_column_list: the width for each column
    :param table_type: type of output (value : 'RST' or 'MD' only)
    :return: header separator
    """
    if table_type == 'RST':
        separator = '+'
        header = '='

    elif table_type == 'MD':
        separator = '|'
        header = '-'
    else:
        raise ValueError('error table type value {} not valid'.format(table_type))

    combi = header + separator + header

    return separator + header + combi.join(
        [header * width_value for width_value in width_column_list]) + header + separator


def _get_line_separator(width_column_list):
    """
    Return a string separator between line of data table
    :param width_column_list:
    :return:
    """
    return '+-' + '-+-'.join(['-' * width_value for width_value in width_column_list]) + '-+'


# TODO add header
def _get_features_data_table_line(
    geolayer,
    field_name_list=None,
    print_i_feat=True,
    table_type='RST',
    light=True,
    display_geo_data=True,
    max_width_coordinates_printing=30,
    limit=None,
    header=True
):
    """
    Returns a generator to display the attribute data of a geolayer
    
    :param geolayer: geolayer that we want print
    :param field_name_list: if filled out we restrain printing on only indicate fields
    :param print_i_feat: choose if you want to see appears i_feat or not (default : NOT)
    :param table_type: choose type of printing ( 2 format allowed Markdown ('MD') or reStructuredText (RST))
    :param light: choose if you want a separator between line or not (default: NOT)
    :param display_geo_data: print geometry information (Geometry Type and Coordinates extract) ? (default: NOT)
    :param max_width_coordinates_printing: if display_geometry is True you can choose the max width of coordinates
           column. (if None all coordinates are print).
    :param header: print table header

    :return: line that compose table
    """
    # init attributes_values and geometry_values
    if 'fields' in geolayer['metadata']:
        attributes_values = True
    else:
        attributes_values = False

    if 'geometry_ref' in geolayer['metadata']:
        geometry_values = True
    else:
        geometry_values = False

    if not field_name_list:
        # create field name
        if 'fields' in geolayer['metadata']:
            attributes_values = True
            field_name_list = [None] * len(geolayer['metadata']['fields'])
            for i_field, field_name in enumerate(geolayer['metadata']['fields']):
                if 'index' in geolayer['metadata']['fields'][field_name]:
                    idx_field = geolayer['metadata']['fields'][field_name]['index']
                else:
                    idx_field = i_field
                field_name_list[idx_field] = field_name
        else:
            field_name_list = []

    if attributes_values:
        max_width_value_list = [len(field_name) for field_name in field_name_list]
    else:
        max_width_value_list = []

    if print_i_feat is True:
        key_max_width = len('i_feat')

    if display_geo_data:
        if "geometry_ref" in geolayer['metadata']:
            geometry_values = True
            max_width_value_list += [len('type'), len('coordinates')]

    # scan geolayer data for determine output width table
    for i_limit, (i_feat, feature) in enumerate(geolayer['features'].items()):
        # if limit we stop loop
        if limit and i_limit == limit:
            break
        # if feature is serialized
        if 'feature_serialize' in geolayer['metadata']:
            if geolayer['metadata']['feature_serialize'] is True:
                feature = eval(feature)

        #  DATA
        # get attributes data
        if attributes_values:
            if 'attributes' in feature:
                width_value_list = [len(str(feature['attributes'][field_name])) if field_name in feature['attributes'] else 0 for field_name in field_name_list]
            else:
                width_value_list = [0] * len(field_name_list)
        else:
            width_value_list = []

        # GEOMETRY
        # get geometry data
        if geometry_values:
            if display_geo_data:
                if 'geometry' in feature:
                    if 'coordinates' in feature['geometry']:
                        coordinates_geometries_key = 'coordinates'
                    elif 'geometries' in feature['geometry']:
                        coordinates_geometries_key = 'geometries'
                    else:
                        raise Exception('unreadable geometry')
                    coordinates = feature['geometry'][coordinates_geometries_key]
                    width_coordinates = len(str(coordinates))

                    if max_width_coordinates_printing:
                        if width_coordinates > max_width_coordinates_printing:
                            width_coordinates = max_width_coordinates_printing

                    width_value_list += [len(str(feature['geometry']['type'])), width_coordinates]
                else:
                    width_value_list += [0, 0]

        # resize width of column if feature's data are larger than other one
        for i_width_value, width_value in enumerate(width_value_list):
            if width_value > max_width_value_list[i_width_value]:
                max_width_value_list[i_width_value] = width_value

            # compute key width
            if print_i_feat is True:
                if len(str(i_feat)) > key_max_width:
                    key_max_width = len(str(i_feat))
    # OPTIONS
    # i_feat
    if print_i_feat is True:
        # add i_feat max wisth to field_name_list
        max_width_value_list = [key_max_width] + max_width_value_list
        # add 'i_feat' value to complete_field_name_list
        complete_field_name_list = ['i_feat'] + field_name_list
    else:
        complete_field_name_list = field_name_list

    # add field type and coordinates at the end
    if display_geo_data:
        if 'geometry_ref' in geolayer["metadata"]:
            complete_field_name_list = complete_field_name_list + ['type', 'coordinates']

    # RETURN LINE TABLE
    # generate header top
    if table_type == 'RST':
        yield _get_line_separator(max_width_value_list)

    # generate header
    yield _get_line(fields_value_list=complete_field_name_list, max_width_value_list=max_width_value_list)
    # generate header seprator
    yield _get_header_separator(max_width_value_list, table_type=table_type)

    # generate line for each value in geolayer
    for i_limit, (i_feat, feature) in enumerate(geolayer['features'].items()):
        # if limit we stop loop
        if limit and i_limit == limit:
            break
        # if feature is serialized
        if 'feature_serialize' in geolayer['metadata']:
            if geolayer['metadata']['feature_serialize'] is True:
                feature = eval(feature)

        # get attributes value
        if attributes_values:
            if 'attributes' in feature:
                field_value_list = [feature['attributes'][field_name] if field_name in feature['attributes'] else None for field_name in field_name_list]
            else:
                field_value_list = [None] * len(geolayer['metadata']['fields'])
        else:
            field_value_list = []
        
        # adding i_feat if option is True
        if print_i_feat is True:
            field_value_list = [i_feat] + field_value_list

        if display_geo_data:
            if "geometry" in feature:
                geometry_type = feature['geometry']['type']
                field_value_list += [geometry_type]

                if geometry_type == 'GeometryCollection':
                    coordinates_value = str(feature['geometry']['geometries'])
                else:
                    coordinates_value = str(feature['geometry']['coordinates'])

                if max_width_coordinates_printing:
                    coordinates_value = coordinates_value[:max_width_coordinates_printing]

                    if len(coordinates_value) == max_width_coordinates_printing:
                        if geometry_type.upper() in ['POINT', 'GEOMETRYCOLLECTION']:  # ['Point', 'GeometryCollection']
                            coordinates_value = coordinates_value[:-5] + ' ...]'
                        elif geometry_type.upper() in ['LINESTRING', 'MULTIPOINT']:  # ['LineString', 'MultiPoint']
                            coordinates_value = coordinates_value[:-6] + ' ...]]'
                        elif geometry_type.upper() in ['POLYGON', 'MULTILINESTRING']:  # ['Polygon', 'MultiLineString']
                            coordinates_value = coordinates_value[:-7] + ' ...]]]'
                        elif geometry_type.upper() in ['MULTIPOLYGON']:  # ['MultiPolygon']
                            coordinates_value = coordinates_value[:-8] + ' ...]]]]'

                field_value_list += [coordinates_value]
            else:
                field_value_list += [None, None]

        # generate feature value
        yield _get_line(field_value_list, max_width_value_list)

        # generate line separator if output format is RST
        if table_type == 'RST':
            if light is False:
                yield _get_line_separator(max_width_value_list)

    # generate line separator if output format is RST
    if table_type == 'RST':
        yield _get_line_separator(max_width_value_list)


def print_features_data_table(
    geolayer,
    field_name_list=None,
    print_i_feat=True,
    table_type='RST',
    light=True,
    display_geometry=True,
    max_width_coordinates_printing=30,
    header=True,
    limit=None
):
    """
    Print geolayer features attributes and/or geometry when exists.

    :param geolayer: geolayer that we want print
    :param field_name_list: if filled out we restrain printing on only indicate fields
    :param print_i_feat: choose if you want to see appears i_feat or not (default : NOT)
    :param table_type: choose type of printing ( 2 format allowed Markdown ('MD') or reStructuredText (RST))
    :param light: choose if you want a separator between line or not (default: NOT)
    :param display_geometry: print geometry information (Geometry Type and Coordinates extract) ? (default: NOT)
    :param max_width_coordinates_printing: if display_geo_data is True you can choose the max width of coordinates
    column. (if None all coordinates are print).
    :param header: print table header
    :param limit: if you want just print x features.

    :return: line that compose table
    """
    return_value = ''
    for i_line, line in enumerate(_get_features_data_table_line(
        geolayer=geolayer,
        field_name_list=field_name_list,
        print_i_feat=print_i_feat,
        table_type=table_type,
        light=light,
        display_geo_data=display_geometry,
        max_width_coordinates_printing=max_width_coordinates_printing,
        header=True,
        limit=limit
    )):
        return_value += line + '\n'

    return return_value


# TODO add header
def _get_fields_metadata_table_line(
    geolayer,
    field_name_list=None,
    key_field_name='field name',
    order_value=False,
    table_type='RST',
    light=True,
    header=True
):
    """
    Returns a generator to display the metadata data of a geolayer

    :param geolayer:
    :param field_name_list: if filled out we restrain printing on only indicate fields
    :param key_field_name: name of column that contains field_name
    :param order_value: respect the order apparition of field_name ('index' key must be filled out in field metadata).
           (default: NOT)
    :param table_type: choose type of printing ( 2 format allowed Markdown ('MD') or reStructuredText (RST))
    :param light: choose if you want a separator between line or not (default: NOT)
    :param header: print table header

    :return: line that compose table
    """

    dict_data = geolayer['metadata'].get('fields')
    if dict_data:
        if not field_name_list:
            field_name_list = ['type', 'width', 'precision', 'index']

        max_width_value_list = [len(str(field_name)) for field_name in field_name_list]

        # create a calling list who is
        call_feature_order = list(dict_data.keys())

        # compute max width for data_value
        key_max_width = 0
        for i_field, key in enumerate(dict_data):
            width_value_list = [len(str(dict_data[key][field_name])) if field_name in dict_data[key] else len('None') for field_name in field_name_list]
            # compute width data_value
            for i_width_value, width_value in enumerate(width_value_list):
                if width_value > max_width_value_list[i_width_value]:
                    max_width_value_list[i_width_value] = width_value
            # compute key width
            if len(str(key)) > key_max_width:
                key_max_width = len(str(key))

            if order_value:
                idx = dict_data[key]['index']
                call_feature_order[idx] = key

        # add max key width to key width value list
        if len(key_field_name) > key_max_width:
            key_max_width  = len(key_field_name)
        max_width_value_list = [key_max_width] + max_width_value_list

        # add key_field_name to field_name_list
        complete_field_name_list = [key_field_name] + field_name_list

        # create line separator
        line_separator = _get_line_separator(max_width_value_list)

        # create fields name line
        line_field_name = _get_line(complete_field_name_list, max_width_value_list)

        # generate table to print
        if table_type == 'RST':
            yield line_separator
        yield line_field_name
        yield _get_header_separator(max_width_value_list, table_type=table_type)

        # for i_field, key in enumerate(dict_data):
        for i_feat in call_feature_order:
            field_value_list = [i_feat] + [dict_data[i_feat][field_name] if field_name in dict_data[i_feat] else None for field_name in field_name_list]
            yield _get_line(field_value_list, max_width_value_list)
            if not light:
                yield line_separator

        if table_type == 'RST':
            yield _get_line_separator(max_width_value_list)


def print_metadata_field_table(
    geolayer,
    field_name_list=None,
    key_field_name='field name',
    order_value=False,
    table_type='RST',
    light=True,
    header=True
):
    """
    Return a string metadata abstract.

    :param geolayer:
    :param field_name_list: if filled out we restrain printing on only indicate fields
    :param key_field_name: name of column that contains field_name
    :param order_value: respect the order apparition of field_name ('index' key must be filled out in field metadata).
    (default: NOT)
    :param table_type: choose type of printing ( 2 format allowed Markdown ('MD') or reStructuredText (RST))
    :param light: choose if you want a separator between line or not (default: NOT)
    :param header: print table header

    :return: line that compose table
    """
    return_value = ''
    for line in _get_fields_metadata_table_line(
        geolayer=geolayer,
        field_name_list=field_name_list,
        key_field_name=key_field_name,
        order_value=order_value,
        table_type=table_type,
        light=light,
        header=header
    ):
        return_value += line + '\n'

    return return_value



import csv
import json

from geoformat.conf.error_messages import geometry_format_not_exists
from geoformat.conf.fields_variable import geoformat_field_type_to_csv_field_type
from geoformat.conf.fields_variable import none_value_pattern
from geoformat.conf.geoformat_var import GEOLAYER_DEFAULT_NAME
from geoformat.conf.path import path_to_file_path
from geoformat.conversion.feature_conversion import feature_filter
from geoformat.manipulation.geolayer_manipulation import feature_list_to_geolayer
from geoformat.conversion.fields_conversion import recast_field_value
from geoformat.conversion.geometry_conversion import (
    geometry_to_wkb,
    geometry_to_wkt,
    wkb_to_geometry,
    wkt_to_geometry,
)
from geoformat.conversion.metadata_conversion import (
    get_field_name_list_ordered_by_i_field,
)
from geoformat.driver.common_driver import _get_recast_field_type_mapping
from geoformat.driver.common_driver import load_data


def _from_csv_get_features_list(
    csv_reader,
    header,
    serialize,
    field_name_filter,
    geometry_field,
    geometry_field_name,
    geometry_format,
    geometry_type_filter,
    bbox_filter,
    bbox
):
    """
    Take csv reader object and yield csv row into geoformat feature

    :param csv_reader: csv reader object.
    :param header: Specifies that the file contains a header line with the names of each column in the file. (bool)

    :return: geoformat Feature
    """
    for i_row, row in enumerate(csv_reader):
        if i_row == 0:
            if header is True:
                field_name_list = list(row)
                continue
            else:
                field_name_list = [
                    "field_{}".format(i_field) for i_field in range(len(row))
                ]

        feature = {}
        feature["attributes"] = {
            field_name: row[i_field]
            for i_field, field_name in enumerate(field_name_list)
        }

        if geometry_field is True:
            geometry_as_attributes = feature["attributes"].get(
                geometry_field_name
            )
            if geometry_format in {"WKT", "WKB", "GEOJSON"}:
                if geometry_format == "WKT":
                    geoformat_geometry = wkt_to_geometry(geometry_as_attributes, bbox=bbox)
                elif geometry_format == "WKB":
                    geoformat_geometry = wkb_to_geometry(geometry_as_attributes, bbox=bbox)
                else:  # GEOJSON
                    geoformat_geometry = json.loads(geometry_as_attributes)
            else:
                raise Exception(geometry_format_not_exists)

            if geoformat_geometry:
                feature["geometry"] = geoformat_geometry
                # delete feature attributes for geometry_field_name
                del feature["attributes"][geometry_field_name]

        feature = feature_filter(
            feature=feature,
            serialize=serialize,
            field_name_filter=field_name_filter,
            geometry_type_filter=geometry_type_filter,
            bbox_filter=bbox_filter,
            bbox=bbox
        )
        if feature:
            yield feature


def csv_to_geolayer(
    path,
    geolayer_name=None,
    delimiter=";",
    header=True,
    null_string='',
    # quote_character='', TODO
    field_name_filter=None,
    force_field_conversion=False,
    serialize=False,
    encoding="utf8",
    geometry_field=False,
    geometry_field_name=None,
    geometry_type_filter=None,
    geometry_format="WKT",
    crs="",
    bbox_extent= False,
    bbox_filter= None,
    http_headers=None
):
    """
    From csv file get a geolayer.

    :param path: path to csv file
    :param delimiter: Specifies the character that separates columns within each row (line) of the file.
     The default is comma character.
    :param header: Specifies that the file contains a header line with the names of each column in the file. (bool)
    :param field_name_filter: filter only on specified field_name (can be a list)
    :param force_field_conversion: True if you want to force value in field (can change field type) / False if you want
           to deduce field type without forcing field type.
    :param serialize: True if features in geolayer are serialized (can reduce performance) / False if not.
    :param geometry_field:
    :param geometry_field_name: field's name that contain geometry
    :param geometry_format: geometry format (WKT / WKB or GEOJSON)
    :param encoding: string encoding default utf8
    :param http_headers: optionally you can add headers of http parameters

    :return: geolayer
    """

    # update none_value_pattern variable
    if null_string is not None:
        none_value_pattern.update([null_string])

    # with open(p, "r", encoding=encoding) as csv_file:
    csv_file, csv_name = load_data(path=path, encoding=encoding, http_headers=http_headers)
    geolayer_name = geolayer_name or csv_name or GEOLAYER_DEFAULT_NAME
    with open(csv_file.name, 'r') as csv_data:
        raw_feature_list = csv.reader(csv_data, delimiter=delimiter)
        raw_feature_list = list(
            _from_csv_get_features_list(
                csv_reader=raw_feature_list,
                header=header,
                serialize=serialize,
                field_name_filter=field_name_filter,
                geometry_field=geometry_field,
                geometry_field_name=geometry_field_name,
                geometry_format=geometry_format,
                geometry_type_filter=geometry_type_filter,
                bbox_filter=bbox_filter,
                bbox=bbox_extent
            )
        )
        geolayer = feature_list_to_geolayer(
            feature_list=raw_feature_list,
            geolayer_name=geolayer_name,
            force_field_conversion=force_field_conversion,
            bbox_extent=bbox_extent,
            crs=crs,
            serialize=serialize,
            none_value_pattern=none_value_pattern,
        )

    return geolayer


def geoformat_feature_to_csv_feature(
    feature,
    recast_field_mapping,
    string_field_name,
    null_string,
    quote_character,
    write_geometry,
    geometry_format,
):
    """
    Make transformation between geoformat feature to geojson feature.

    :param feature: geoformat feature
    :param recast_field_mapping: dict that contains field to recast and parameters to recast
    :param string_field_name: field's name
    :param null_string: how to format in csv format None value
    :param quote_character: quote character in csv to describe string value
    :param write_geometry: True if you want write geometry field in csv
    :param geometry_format: convert geometry to output format

    :return: dict like geoformat feature but with attributes and geometry in csv format
    """
    geometry = {}
    if write_geometry is True:
        # get geometries
        geometry = dict(feature.get("geometry", {}))
        if geometry:
            # delete bbox if exists
            geometry.pop("bbox", None)
            if geometry_format == "WKT":
                geometry = geometry_to_wkt(geometry)
            elif geometry_format == "WKB":
                geometry = geometry_to_wkb(geometry).hex()
            elif geometry_format == "GEOJSON":
                geometry = json.dumps(geometry)
            else:
                raise Exception(geometry_format_not_exists)

    # get attributes and clean it
    attributes = dict(feature.get("attributes", {}))
    attributes = feature_attributes_to_csv_attributes(attributes, recast_field_mapping)

    for field_name, value in attributes.items():
        if field_name in string_field_name:
            attributes[field_name] = "{}{}{}".format(
                quote_character, value, quote_character
            )
        if value is None:
            attributes[field_name] = null_string

    return {"attributes": attributes, "geometry": geometry}


def feature_attributes_to_csv_attributes(attributes, recast_field_mapping=None):
    """
    Transforms feature's attributes to geojson properties

    :param attributes: feature attributes
    :param recast_field_mapping: config dict to recast field to geojson properties format
    :return: formatted geojson properties
    """
    if attributes:
        # check is data are serialized or not
        csv_attributes = dict(attributes)
        # if recast_field_mapping
        if recast_field_mapping is not None:
            for field_name, field_name_mapping in recast_field_mapping.items():
                field_value_to_recast = attributes.get(field_name, None)
                if field_value_to_recast is not None:
                    field_value_recast = recast_field_value(
                        field_value=field_value_to_recast, **field_name_mapping
                    )
                    csv_attributes[field_name] = field_value_recast
    else:
        csv_attributes = {}

    return csv_attributes


def geolayer_to_csv(
    geolayer,
    path,
    overwrite=False,
    add_extension=False,
    delimiter=";",  # we prefer semicolon because StringList, IntegerList and FloatList
    header=True,
    null_string="",
    quote_character="",
    write_geometry=True,
    geometry_format="WKT",
    geometry_field_name="geom",
    encoding=None,
):
    """
        Save geolayer to geojson format

    :param geolayer: geolayer that we want to convert in csv,
    :param path: path to output csv,
    :param overwrite: if path exists True to overwrite it,
    :param add_extension: add .csv extension if not exists in path,
    :param delimiter: delimiter character,
    :param header: True to add header in csv
    :param null_string: format of None value in csv,
    :param quote_character: quote character in csv to describe string value,
    :param write_geometry: True to add feature geometry in csv False not,
    :param geometry_format: geometry export format (WKT or WKB or GEOJSON),
    :param geometry_field_name: name of geometry field,
    :param encoding: string encoding default utf8

    """
    # check input path and deduce output path
    output_path = path_to_file_path(
        path=path,
        geolayer_name=geolayer["metadata"]["name"],
        overwrite=overwrite,
        add_extension=add_extension,
    )

    with open(output_path, "w", encoding=encoding) as csv_file:
        string_field_name = []
        geolayer_fields_metadata = geolayer["metadata"].get("fields")
        if geolayer_fields_metadata:
            # get field_name list ordered by i_field
            field_name_list = get_field_name_list_ordered_by_i_field(
                geolayer_fields_metadata
            )
            string_field_name = [
                field_name
                for field_name in geolayer_fields_metadata
                if geolayer_fields_metadata[field_name]["type"] == "String"
            ]

        header_line = []
        if header is True:
            if geolayer_fields_metadata:
                header_line = delimiter.join(
                    [field_name for field_name in field_name_list]
                )
            if write_geometry and geolayer["metadata"].get("geometry_ref"):
                if header_line:
                    header_line = delimiter.join([header_line, geometry_field_name])
                else:
                    header_line = geometry_field_name
            header_line = [header_line]

        # scan field metadata to check if field must have to be recast
        recast_field_mapping = _get_recast_field_type_mapping(
            fields_metadata=geolayer_fields_metadata,
            geoformat_type_to_output_driver_type=geoformat_field_type_to_csv_field_type,
        )
        features_lines = [None] * len(geolayer["features"])
        for i_line, (i_feat, feature) in enumerate(geolayer["features"].items()):
            csv_feature = geoformat_feature_to_csv_feature(
                feature=feature,
                recast_field_mapping=recast_field_mapping,
                string_field_name=string_field_name,
                null_string=null_string,
                quote_character=quote_character,
                write_geometry=write_geometry,
                geometry_format=geometry_format,
            )

            # format attributes
            csv_attributes = csv_feature.get("attributes", [])
            csv_feature_line = None
            if csv_attributes:
                csv_feature_line = delimiter.join(
                    [
                        csv_attributes.get(field_name, null_string)
                        for field_name in field_name_list
                    ]
                )

            # format geometry
            if write_geometry is True:
                feature_geometry = csv_feature.get("geometry", "")
                if feature_geometry:
                    # add geometry
                    if csv_feature_line:
                        csv_feature_line = delimiter.join(
                            [csv_feature_line, feature_geometry]
                        )
                    else:
                        csv_feature_line = feature_geometry

            features_lines[i_line] = csv_feature_line

        # format and write
        csv_lines = header_line + features_lines
        csv_txt = "\n".join(csv_lines)
        csv_file.write(csv_txt)



try:
    from osgeo import ogr
    from osgeo import osr
    import_ogr_success = True
except ImportError:
    import_ogr_success = False

from geoformat.driver.ogr.ogr_driver import ogr_layer_to_geolayer

from geoformat.conf.error_messages import import_ogr_error

def sql(
        sql_request,
        host,
        database_name,
        user,
        password,
        port=5432,
    ):
    """
    Execute an SQL request on the specified PostgreSQL database using GDAL/OGR.

    :param sql_request: SQL request as a string.
    :param host: Host address of the PostgreSQL server.
    :param database_name: Name of the database.
    :param user: Username for the database connection.
    :param password: Password for the database connection.
    :param port: Port number for the database connection (default is 5432).
    :return: Result of the SQL request execution from OGR.
    :raises Exception: If OGR import fails.
    """
    pg_adress = f"PG: host={host} dbname={database_name} user={user} password={password} port={port}"
    if import_ogr_success is True:
        data_source = ogr.Open(pg_adress)
        return data_source.ExecuteSQL(sql_request)
    else:
        raise Exception(import_ogr_error)


def sql_select_to_geolayer(
    select_request,
    host,
    database_name,
    user,
    password,
    port=5432,
    geolayer_name=None,
    field_name_filter=None,
    bbox_extent=True,
    bbox_filter=None,
    feature_serialize=False,
    feature_limit=None,
    feature_offset=None
):
    """
    Execute a SQL SELECT request on the specified PostgreSQL database and return the result as a geolayer.

    :param select_request: SQL SELECT request as a string.
    :param host: Host address of the PostgreSQL server.
    :param database_name: Name of the database.
    :param user: Username for the database connection.
    :param password: Password for the database connection.
    :param port: Port number for the database connection (default is 5432).
    :param geolayer_name: Optional name for the resulting geolayer.
    :param field_name_filter: Optional filter for field names.
    :param bbox_extent: Boolean indicating whether to use bounding box extent (default is True).
    :param bbox_filter: Optional bounding box filter.
    :param feature_serialize: Boolean indicating whether to serialize features (default is False).
    :param feature_limit: Optional limit for the number of features.
    :param feature_offset: Optional offset for the features.
    :return: A geolayer resulting from the given SQL SELECT request.
    """
    # drop view if exists
    sql_drop_view = """DROP VIEW IF EXISTS geoformat_temporary_view;"""
    sql(sql_drop_view, host, database_name, user, password, port)
    # create view request
    sql_create_view = f"""CREATE OR REPLACE VIEW geoformat_temporary_view AS (
    {select_request}
    );"""
    # execute request
    sql(sql_create_view, host, database_name, user, password, port)

    pg_adress = f"PG: host={host} dbname={database_name} user={user} password={password} port={port}"
    geolayer = ogr_layer_to_geolayer(
        pg_adress,
        layer_id_or_name='geoformat_temporary_view',
        field_name_filter=field_name_filter,
        bbox_extent=bbox_extent,
        bbox_filter=bbox_filter,
        serialize=feature_serialize,
        feature_limit=feature_limit,
        feature_offset=feature_offset
    )

    if geolayer_name:
        geolayer['metadata']['name'] = geolayer_name
    # drop the view
    sql_drop_view = """DROP VIEW geoformat_temporary_view;"""
    sql(sql_drop_view, host, database_name, user, password, port)

    return geolayer



import struct

big_endian_wkb_geometry_type_to_geometry_type = {
    b'\x00\x00\x00\x01': 'Point',
    b'\x00\x00\x00\x02': 'LineString',
    b'\x00\x00\x00\x03': 'Polygon',
    b'\x00\x00\x00\x04': 'MultiPoint',
    b'\x00\x00\x00\x05': 'MultiLineString',
    b'\x00\x00\x00\x06': 'MultiPolygon',
    b'\x00\x00\x00\x07': 'GeometryCollection',
}

geometry_type_to_wkb_geometry_type = {
    'POINT': b'\x00\x00\x00\x01',
    'LINESTRING': b'\x00\x00\x00\x02',
    'POLYGON': b'\x00\x00\x00\x03',
    'MULTIPOINT': b'\x00\x00\x00\x04',
    'MULTILINESTRING': b'\x00\x00\x00\x05',
    'MULTIPOLYGON': b'\x00\x00\x00\x06',
    'GEOMETRYCOLLECTION': b'\x00\x00\x00\x07',
}


def int_to_4_bytes_integer(integer_value, integer_endian_big=True):
    """
    make translation between int type value to 4 bytes value.
    Optionally you can choose output bytes endian

    :param integer_value: integer value
    :param integer_endian_big: True if output big endian / False if output little endian
    :return: input value in 4 bytes
    """
    if integer_endian_big is True:
        struct_format = ">i"
    else:
        struct_format = "<i"

    return bytearray(struct.pack(struct_format, integer_value))


def float_to_double_8_bytes_array(float_value, float_big_endian=True):
    """
    Make translation between float type value to 8 bytes value.
    Optionally you can choose output bytes endian

    :param float_value: float value
    :param float_big_endian: True if output big endian / False if output little endian
    :return: input value in 8 bytes
    """
    if float_big_endian is True:
        struct_format = ">d"
    else:
        struct_format = "<d"

    return bytearray(struct.pack(struct_format, float_value))


def coordinates_list_to_bytes(coordinates_list, coordinates_big_endian=True):
    """
    Transform a list of coordinates to bytes value.
    Optionally you can choose the output endian.

    :param coordinates_list: list of coordinates.
    :param coordinates_big_endian: True if output big endian / False if output little endian.
    :return: coordinates in bytes value.
    """

    def coordinates_to_bytes(coordinates, float_big_endian=True):
        """
        Transform coordinates to bytes.
        Optionally you can choose the output endian.

        :param coordinates:
        :param float_big_endian:
        :return: coordinates in bytes value.
        """
        coordinates_bytes = bytearray()
        for coordinate in coordinates:
            coordinate_bytes = float_to_double_8_bytes_array(float_value=coordinate, float_big_endian=float_big_endian)
            coordinates_bytes += coordinate_bytes

        return coordinates_bytes

    first_coord = coordinates_list[0]
    if isinstance(first_coord, (list, tuple)):
        bytes_coordinates = int_to_4_bytes_integer(len(coordinates_list), integer_endian_big=coordinates_big_endian)
        for coord in coordinates_list:
            bytes_coordinates += coordinates_list_to_bytes(coordinates_list=coord,
                                                           coordinates_big_endian=coordinates_big_endian)
    elif isinstance(first_coord, (float, int)):
        bytes_coordinates = coordinates_to_bytes(coordinates=coordinates_list,
                                                 float_big_endian=coordinates_big_endian)
    else:
        raise ValueError

    return bytes_coordinates


def double_8_bytes_to_float(double_8_bytes, double_big_endian, double_dimension=2):
    """
    Make translation between dimension * 8 bytes value to float type value.
    You have to give the endian of data and dimension of coordinates.

    :param double_8_bytes: coordinates value in 8 bytes
    :param double_big_endian: the endian order of your bytes
    :param double_dimension: number of dimension (default 2) for your coordinates
    :return: float tuple of coordinates
    """
    if double_big_endian is True:
        struct_format = ">{dim}d".format(dim=double_dimension)
    elif double_big_endian is False:
        struct_format = "<{dim}d".format(dim=double_dimension)
    else:
        raise ValueError('endian_big must be a bool type')

    return struct.unpack(struct_format, double_8_bytes)


def integer_4_bytes_to_int(integer_4_bytes, integer_endian_big):
    """
    Make translation between integer 4 bytes value to integer type value.
    You have to give the endian of data and double_dimension of coordinates.

    :param integer_4_bytes: coordinates value in 8 bytes
    :param integer_endian_big: the endian of your bytes
    :return: float tuple of coordinates
    """
    if integer_endian_big is True:
        struct_format = ">i"
    elif integer_endian_big is False:
        struct_format = "<i"
    else:
        raise ValueError('endian_big must be a bool type')

    return struct.unpack(struct_format, integer_4_bytes)[0]



from __future__ import annotations

import os
import sys

from geoformat.conf.error_messages import import_matplotlib_error, import_numpy_error
from geoformat.conversion.coordinates_conversion import force_rhr_polygon_coordinates
from geoformat.conversion.geometry_conversion import geometry_to_bbox
from geoformat.geoprocessing.merge_geometries import merge_geometries

# ---------------------------------------------------------------------
# Choose an interactive backend if a GUI is available
# ---------------------------------------------------------------------
import matplotlib


def _environment_supports_gui() -> bool:
    """Return True if a graphical display seems available."""
    return bool(
        os.environ.get("DISPLAY")
        or os.environ.get("WAYLAND_DISPLAY")
        or sys.platform.startswith("win")
    )


def _ensure_interactive_backend() -> None:
    """
    If no MPLBACKEND is set and a GUI is available, try to use an interactive
    backend (QtAgg → Qt5Agg → TkAgg).
    """
    if os.environ.get("MPLBACKEND") or not _environment_supports_gui():
        return
    for candidate in ("QtAgg", "Qt5Agg", "TkAgg"):
        try:
            matplotlib.use(candidate, force=True)
            break
        except Exception:
            continue


_ensure_interactive_backend()

# ---------------------------------------------------------------------
# Imports depending on the environment
# ---------------------------------------------------------------------
try:
    import numpy as np
    import_numpy_success = True
except ImportError:
    import_numpy_success = False

try:
    import matplotlib.pyplot as plt
    from matplotlib.patches import PathPatch
    from matplotlib.path import Path
    import_matplotlib_success = True
except ImportError:
    import_matplotlib_success = False


# ---------------------------------------------------------------------
# Implementation
# ---------------------------------------------------------------------
if import_matplotlib_success and import_numpy_success:

    class DrawGeometry:
        """
        Utility to draw GeoJSON-like geometries with Matplotlib.
        """

        # ---- default style ----
        POINT_MARKER = "o"
        POINT_MARKER_SIZE = 10
        POINT_COLOR = "black"
        POINT_LINESTYLE = "None"
        POINT_Z_ORDER = 2

        LINESTRING_COLOR = "black"
        LINESTRING_FACECOLOR = "none"
        LINESTRING_WIDTH = 1
        LINESTRING_LINESTYLE = "-"
        LINESTRING_Z_ORDER = 1

        POLYGON_EDGE_COLOR = "black"
        POLYGON_FACE_COLOR = "#d8e0ea"
        POLYGON_WIDTH = 1
        POLYGON_LINESTYLE = "-"
        POLYGON_Z_ORDER = 0

        GRATICULE_MAJOR_COLOR = "black"
        GRATICULE_MAJOR_LINE_STYLE = "-"
        GRATICULE_MAJOR_LINE_WIDTH = 0.5
        GRATICULE_MAJOR_ALPHA = 0.5
        GRATICULE_MINOR_COLOR = "grey"
        GRATICULE_MINOR_LINE_STYLE = ":"
        GRATICULE_MINOR_LINE_WIDTH = 0.2
        GRATICULE_MINOR_ALPHA = 0.5

        VERTEX_MARKER = "x"
        VERTEX_MARKER_SIZE = 6
        VERTEX_COLOR = "black"
        VERTEX_LINESTYLE = "None"

        def __init__(self, geometry: dict, *, show_vertices: bool = False) -> None:
            """
            Parameters
            ----------
            geometry : dict
                GeoJSON-like geometry.
            show_vertices : bool, optional
                If True, draw all vertices as markers.
            """
            self.geometry = geometry
            self.bbox = geometry.get("bbox") or geometry_to_bbox(geometry=self.geometry)
            self.fig, self.ax = plt.subplots()
            self.show_vertices = show_vertices

        # ---- helpers ----
        def create_codes(self, num_points: int) -> list[int]:
            return [Path.MOVETO] + [Path.LINETO] * (num_points - 1)

        def validate_coordinates(self, coordinates: list) -> bool:
            return bool(coordinates and (not isinstance(coordinates[0], (list, tuple)) or any(coordinates)))

        def _plot_vertices(self, points: list[list[float]]) -> None:
            if not points:
                return
            arr = np.array(points)
            self.ax.plot(
                arr[:, 0],
                arr[:, 1],
                marker=self.VERTEX_MARKER,
                markersize=self.VERTEX_MARKER_SIZE,
                color=self.VERTEX_COLOR,
                linestyle=self.VERTEX_LINESTYLE,
                zorder=3,
            )

        # ---- plot primitives ----
        def plot_point(self, coordinates: list[float]) -> None:
            self.ax.plot(
                coordinates[0],
                coordinates[1],
                marker=self.POINT_MARKER,
                markersize=self.POINT_MARKER_SIZE,
                color=self.POINT_COLOR,
                linestyle=self.POINT_LINESTYLE,
                zorder=self.POINT_Z_ORDER,
            )
            if self.show_vertices:
                self._plot_vertices([coordinates])

        def plot_line_string(self, coordinates: list[list[float]]) -> None:
            verts = np.array(coordinates)
            path = Path(verts)
            patch = PathPatch(
                path,
                edgecolor=self.LINESTRING_COLOR,
                facecolor=self.LINESTRING_FACECOLOR,
                linewidth=self.LINESTRING_WIDTH,
                linestyle=self.LINESTRING_LINESTYLE,
                zorder=self.LINESTRING_Z_ORDER,
            )
            self.ax.add_patch(patch)
            if self.show_vertices:
                self._plot_vertices(coordinates)

        def plot_polygon(self, coordinates: list[list[list[float]]]) -> None:
            coordinates = force_rhr_polygon_coordinates(coordinates=coordinates)
            verts = None
            codes: list[int] = []
            for ring in coordinates:
                ring_verts = np.array(ring)
                verts = ring_verts if verts is None else np.concatenate([verts, ring_verts])
                codes += self.create_codes(len(ring))
            path = Path(verts, codes)
            patch = PathPatch(
                path,
                edgecolor=self.POLYGON_EDGE_COLOR,
                facecolor=self.POLYGON_FACE_COLOR,
                linewidth=self.POLYGON_WIDTH,
                linestyle=self.POLYGON_LINESTYLE,
                zorder=self.POLYGON_Z_ORDER,
            )
            self.ax.add_patch(patch)
            if self.show_vertices:
                for ring in coordinates:
                    self._plot_vertices(ring)

        def plot_multi_point(self, coordinates: list[list[float]]) -> None:
            arr = np.array(coordinates)
            self.ax.plot(
                arr[:, 0],
                arr[:, 1],
                marker=self.POINT_MARKER,
                markersize=self.POINT_MARKER_SIZE,
                color=self.POINT_COLOR,
                linestyle=self.POINT_LINESTYLE,
                zorder=self.POINT_Z_ORDER,
            )
            if self.show_vertices:
                self._plot_vertices(coordinates)

        def plot_multi_line_string(self, coordinates: list[list[list[float]]]) -> None:
            for ls in coordinates:
                if ls:
                    self.plot_line_string(ls)

        def plot_multi_polygon(self, coordinates: list[list[list[list[float]]]]) -> None:
            for poly in coordinates:
                if poly:
                    self.plot_polygon(poly)

        # ---- dispatch and final rendering ----
        def plot(
            self,
            graticule: bool = False,
            *,
            save_path: str | os.PathLike[str] | None = None,
            dpi: int | float | None = None,
        ) -> None:
            """
            Render the geometry. Save to file if save_path is provided.

            If save_path is None, try to show the figure interactively.
            In a headless environment without save_path, a RuntimeError is raised.
            """
            if self.geometry["type"] == "GeometryCollection":
                for geometry in self.geometry["geometries"]:
                    self.plot_geometry(geometry)
            else:
                self.plot_geometry(self.geometry)

            margin = self.expand_bbox()
            self.ax.set_xlim(margin[0], margin[2])
            self.ax.set_ylim(margin[1], margin[3])

            if graticule:
                self.ax.minorticks_on()
                self.ax.grid(
                    which="major",
                    color=self.GRATICULE_MAJOR_COLOR,
                    linestyle=self.GRATICULE_MAJOR_LINE_STYLE,
                    linewidth=self.GRATICULE_MAJOR_LINE_WIDTH,
                    alpha=self.GRATICULE_MAJOR_ALPHA,
                )
                self.ax.grid(
                    which="minor",
                    color=self.GRATICULE_MINOR_COLOR,
                    linestyle=self.GRATICULE_MINOR_LINE_STYLE,
                    linewidth=self.GRATICULE_MINOR_LINE_WIDTH,
                    alpha=self.GRATICULE_MINOR_ALPHA,
                )

            self.ax.set_aspect("equal", adjustable="box")

            if save_path:
                plt.savefig(save_path, dpi=dpi)
                return  # no interactive show when saving only

            try:
                plt.show()
            except Exception as e:
                raise RuntimeError(
                    "Interactive display is not available. "
                    "Provide `path='output.png'` to save the figure instead."
                ) from e

        def plot_geometry(self, geometry: dict) -> None:
            handlers = {
                "Point": self.plot_point,
                "MultiPoint": self.plot_multi_point,
                "LineString": self.plot_line_string,
                "MultiLineString": self.plot_multi_line_string,
                "Polygon": self.plot_polygon,
                "MultiPolygon": self.plot_multi_polygon,
            }
            if (handler := handlers.get(geometry["type"])) is not None:
                coords = geometry.get("coordinates", [])
                if self.validate_coordinates(coords):
                    handler(coords)

        def expand_bbox(self) -> tuple[float, float, float, float]:
            if self.bbox:
                x_diff = self.bbox[2] - self.bbox[0]
                y_diff = self.bbox[3] - self.bbox[1]
                x_margin = x_diff * 0.1 or 1
                y_margin = y_diff * 0.1 or 1
                return (
                    self.bbox[0] - x_margin,
                    self.bbox[1] - y_margin,
                    self.bbox[2] + x_margin,
                    self.bbox[3] + y_margin,
                )
            return (-1, -1, 1, 1)


# ---------------------------------------------------------------------
# Public wrapper functions
# ---------------------------------------------------------------------
def draw_geometry(
    geometry: dict,
    graticule: bool = False,
    *,
    show_vertices: bool = False,
    path: str | os.PathLike[str] | None = None,
    dpi: int | float | None = None,
) -> None:
    """
    Draw a geometry. If `path` is given, the figure is saved to that file and not shown.
    """
    if import_matplotlib_success and import_numpy_success:
        DrawGeometry(geometry, show_vertices=show_vertices).plot(
            graticule=graticule, save_path=path, dpi=dpi
        )
    else:
        if not import_matplotlib_success:
            raise Exception(import_matplotlib_error)
        if not import_numpy_success:
            raise Exception(import_numpy_error)


def draw_feature(
    feature: dict,
    graticule: bool = False,
    *,
    show_vertices: bool = False,
    path: str | os.PathLike[str] | None = None,
    dpi: int | float | None = None,
) -> None:
    """
    Draw a GeoJSON-like feature (uses its "geometry").
    If `path` is given, the figure is saved to that file and not shown.
    """
    if import_matplotlib_success and import_numpy_success:
        if geometry := feature.get("geometry"):
            draw_geometry(
                geometry,
                graticule=graticule,
                show_vertices=show_vertices,
                path=path,
                dpi=dpi,
            )
    else:
        if not import_matplotlib_success:
            raise Exception(import_matplotlib_error)
        if not import_numpy_success:
            raise Exception(import_numpy_error)


def draw_geolayer(
    geolayer: dict,
    graticule: bool = False,
    *,
    show_vertices: bool = False,
    path: str | os.PathLike[str] | None = None,
    dpi: int | float | None = None,
) -> None:
    """
    Draw a geolayer by merging all its features' geometries.
    If `path` is given, the figure is saved to that file and not shown.
    """
    if import_matplotlib_success and import_numpy_success:
        geolayer_geometry = None
        features = geolayer.get("features")
        if isinstance(features, dict):
            for _, feature in features.items():
                if geom := feature.get("geometry"):
                    geolayer_geometry = (
                        geom if geolayer_geometry is None else merge_geometries(geolayer_geometry, geom)
                    )
        if geolayer_geometry:
            draw_geometry(
                geolayer_geometry,
                graticule=graticule,
                show_vertices=show_vertices,
                path=path,
                dpi=dpi,
            )
    else:
        if not import_matplotlib_success:
            raise Exception(import_matplotlib_error)
        if not import_numpy_success:
            raise Exception(import_numpy_error)



from importlib.metadata import (
    version as lib_version,
    PackageNotFoundError,
)

def get_version():
    package_name = 'geoformat'
    try:
        return lib_version(package_name)
    except PackageNotFoundError:
        return "unknown version"

__version__ = get_version()


from geoformat.geoprocessing.measure.mesure_area import triangle_area

# - récupère une liste de coordonnées(points) en mètres.
# - récupère une tolérance (superficie minimale à afficher).

# - calcule l'aire du triangle en partant du 1er point (points 1,2,3):
#   (hauteur X base) / 2.
# - change le point de départ (+1) et calcul l'aire du nouveau triangle.
# - stocke en mémoire tous les aires des triangles.
# - s'arrête de calculer les aires dès qu'il n'y a que deux points de disponible.

# - cherche la plus petite aire, supprime le deuxième point du triangle et le nombre de points restant.
# - même chose jusqu'à ce qu'il n'y ait plus rien à simplifier.


def _recompute_idx_point_with_low_area(_coordinate_list, _idx, _coordinates_idx_with_low_area, _tolerance):
    """
    The aim of this function is to recompute, after a point deletion, triangular area for points after and before
    deleted point if it's possible.
    If area if lower than tolerance we add idx point in coordinates_idx_with_low_area to make deletion.


    :param _coordinate_list:
    :param _idx:
    :param _coordinates_idx_with_low_area:
    :param _tolerance:
    :return: updated coordinates_idx_with_low_area
    """

    # TODO rendre possible l'ordre de suppression des points par poids (taille de l'aire) ?
    # TODO usage des sets VS listes (perf-test)

    # if there are enough coordinates to compute triangle after
    # AND we have not yet decide to delete point before
    if _idx >= 2:
        triangle_before_area = triangle_area(vertex_a=_coordinate_list[_idx - 2], vertex_b=_coordinate_list[_idx - 1],
                                             vertex_c=_coordinate_list[_idx])
        if _idx - 1 not in _coordinates_idx_with_low_area:
            if triangle_before_area < _tolerance:
                _coordinates_idx_with_low_area = _coordinates_idx_with_low_area + [_idx - 1]
        # if point immediately before is in _coordinates_idx_with_low_area we have to recompute new area a this point
        # (because area have change) if new area > _tolerance this before point must be keep in coordinate list
        else:
            if triangle_before_area > _tolerance:
                del _coordinates_idx_with_low_area[_coordinates_idx_with_low_area.index(_idx - 1)]

    # if there are enough coordinates to compute triangle after
    if len(_coordinate_list) - 1 >= _idx + 1:
        triangle_after_area = triangle_area(vertex_a=_coordinate_list[_idx - 1], vertex_b=_coordinate_list[_idx],
                                            vertex_c=_coordinate_list[_idx + 1])
        if triangle_after_area < _tolerance:
            _coordinates_idx_with_low_area = _coordinates_idx_with_low_area + [_idx]

    return _coordinates_idx_with_low_area


def visvalingam_whyatt(coordinate_list, tolerance):
    """
    This function is a generalization algorithm.
    It's take a coordinate's list and a tolerance value and return a simplified coordinate's list.

    doc :
        https://en.wikipedia.org/wiki/Visvalingam%E2%80%93Whyatt_algorithm
        https://bost.ocks.org/mike/simplify/


    :param coordinate_list: coordinate's list
    :param tolerance: minimum area in coordinates units
    :return: simplified coordinates
    """
    # if coordinates_list is not empty and have more than 2 coordinates
    if len(coordinate_list) > 2:

        coordinates_idx_with_low_area = []
        nb_coordinates = len(coordinate_list)
        for i_point, point in enumerate(coordinate_list):
            # get points to compute triangular area
            if i_point == 0:
                continue
            elif i_point == nb_coordinates - 1:
                break
            else:
                point_a = coordinate_list[i_point - 1]

            point_b = point
            point_c = coordinate_list[i_point + 1]
            # area calculation
            area = triangle_area(point_a, point_b, point_c)
            if area < tolerance:
                coordinates_idx_with_low_area.append(i_point)

        # if there is coordinates to delete
        if coordinates_idx_with_low_area:
            # we loop on list until there is coordinates inside
            while coordinates_idx_with_low_area:
                # get idx to coordinates and delete it from while coordinates_idx_with_low_area
                idx = coordinates_idx_with_low_area[-1]
                del coordinates_idx_with_low_area[-1]
                # delete low area coordinates with idx
                del coordinate_list[idx]

                # now we have to recalculated area for triangle before and after deleting point (if it's possible)
                coordinates_idx_with_low_area = _recompute_idx_point_with_low_area(
                    _coordinate_list=coordinate_list,
                    _idx=idx,
                    _coordinates_idx_with_low_area=coordinates_idx_with_low_area,
                    _tolerance=tolerance
                )

    return coordinate_list



from geoformat.geoprocessing.connectors.operations import (
    coordinates_to_point,
    coordinates_to_segment
)
from geoformat.geoprocessing.connectors.predicates import (
    point_intersects_point,
    point_intersects_segment
)

from geoformat.conversion.geometry_conversion import (
    single_geometry_to_multi_geometry,
    multi_geometry_to_single_geometry
)
from geoformat.conversion.segment_conversion import (
    segment_list_to_linestring
)

from geoformat.geoprocessing.geoparameters.bbox import bbox_union


def segment_split_by_point(segment, point, tolerance=None):
    """
    This function return a generator with split segments if point intersects it. If not the full segment is yield

    :param segment: segment that you want split with a point.
    :param point: point that will split segment.
    :param tolerance: distance below which two geographic objects are considered to be confused.
    :return: generator with split segment.
    """
    # if point intersects segment
    if point_intersects_segment(point, segment, tolerance):
        point_begin_segment = segment[0]
        point_end_segment = segment[1]
        # test if point is not at segment extremity
        if not point_intersects_point(point, point_begin_segment, tolerance) \
                and not point_intersects_point(point, point_end_segment, tolerance):
            segment_a = point_begin_segment, point
            segment_b = point, point_end_segment
            yield segment_a
            yield segment_b


def linestring_split_by_point(linestring, point, tolerance=None, bbox=True):
    # TODO 1/ add tolerance consideration [DONE]
    # TODO 2/ delete duplicate points (if exists)
    # TODO 3/ optimize filter point intersection bbox_linestring
    # TODO 4/ optimize with segment that intersects points (grid index ?)
    # TODO 5/ manage bbox [DONE]
    # ToDO 6/ if output same that output => return geometry collection ?

    # transform point and linestring to multipoint and multilinestring
    multilinestring = single_geometry_to_multi_geometry(linestring)
    multi_point = single_geometry_to_multi_geometry(point)
    for _point in coordinates_to_point(multi_point['coordinates']):
        segment_split_list = []
        for _linestring in multi_geometry_to_single_geometry(multilinestring):
            # create a separation for new linestring
            segment_split_list.append([])
            # loop on each linestring segments
            for segment in coordinates_to_segment(_linestring['coordinates']):
                if point_intersects_segment(point=_point, segment=segment, tolerance=tolerance):
                    # point to cut can be :
                    #   1. on segment (we split the segment and create a cutting separation in segment_split_list )
                    #   2. at extremity point of segment (segment is preserve but a cutting separation must be done with
                    #   next segment)
                    point_at_extremity = True
                    for i_split, segment_split in enumerate(segment_split_by_point(segment=segment, point=_point, tolerance=tolerance)):
                        # if return split segment we validate condition 1.
                        point_at_extremity = False
                        segment_split_list[-1].append(segment_split)
                        if i_split == 0:
                            segment_split_list.append([])
                    # if point_at_extremity we validate condition 2.
                    if point_at_extremity is True:
                        # test if there is data in segment_split_list
                        if segment_split_list[-1]:
                            # if first point to segment is equal to last point to last segment in segment_split_list
                            # we create a new split
                            if point_intersects_point(_point, segment[0], tolerance) and segment[0] == segment_split_list[-1][-1][1]:
                                segment_split_list.append([])
                        segment_split_list[-1].append(segment)
                else:
                    segment_split_list[-1].append(segment)

        # rebuild new multilinestring with cut linestring
        multilinestring = {'type': 'MultiLinestring', 'coordinates': []}
        for segment_list in segment_split_list:
            __linestring = segment_list_to_linestring(segment_list, bbox=False, segment_as_part=False)
            multilinestring['coordinates'].append(__linestring['coordinates'])

    # prepare return
    geometries_to_return = []
    # if bbox is True
    if bbox is True:
        collection_bbox = ()
    for ___linestring in multi_geometry_to_single_geometry(multilinestring, bbox=bbox):
        if bbox is True and 'bbox' in ___linestring:
            if collection_bbox:
                collection_bbox = bbox_union(collection_bbox, ___linestring['bbox'])
            else:
                collection_bbox = ___linestring['bbox']
        geometries_to_return.append(___linestring)

    output_geometry = {'type': 'GeometryCollection', 'geometries': geometries_to_return}

    if bbox is True:
        output_geometry["bbox"] = collection_bbox

    return output_geometry
